/*
  Example illustrating the use of the BlockIP library.
  It formulates a small multicommodity flow instance,
  which is solved several times with different options.
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <cstring>
#include "math.h"
#include "BlockIP.h"
#include "ExceptionBlockIP.h"

using namespace std;

struct Param_fobj {
  int nBlocks, nNodes, nArcs;
  double *c, *q;
};

void nonlin_objective_function(int n_vars, double x[], double &fx, double Gx[], double Hx[], void *params_) 
{
  Param_fobj *params= (Param_fobj *)params_;
  
  int ncom= params->nBlocks;
  int nnod= params->nNodes;
  int narc= params->nArcs;
  double *c = params->c;
  double *q = params->q;

  fx= 0.0;

  // flows
  for (int i=0; i< ncom*narc; i++) {
    fx += (c[i]+q[i]*x[i])*x[i];
    Gx[i]= c[i]+2.0*q[i]*x[i];
    Hx[i]= 2.0*q[i];
  } 

  // slacks of linking constraints
  for (int i= ncom*narc; i< ncom*narc+narc; i++) {
    Gx[i]= 0.0;
    Hx[i]= 0.0;
  }
}


int main(int argc, char *argv[]) {

  const int numBlocks = 2;  // number of commodities
  const int numNodes = 4; // number of nodes 
  const int numArcs = 5;  // number of arcs
  
  int src[numArcs] = {0, 0, 0, 1, 2}; // source node for each arc
  int dst[numArcs] = {1, 2, 3, 3, 3}; // destination node for each arc
  
  // node injections per commodity
  double injection[numBlocks*numNodes] = {18, 0, 0, -18,
                                          20, 0, 0, -20}; 
  // linear costs per commodity
  double c[numBlocks*numArcs] = {1, 1, 1, 1, 1,
				 2, 2, 2, 2, 2}; 

  // quadratic costs per commodity
  double q[numBlocks*numArcs] = {1, 1, 1, 1, 1,
				 1, 1, 1, 1, 1}; 

  // arc capacities for each commodity
  double ind_cap[numBlocks*numArcs] = {10, 10, 10, 10, 10,
                                       10, 10, 10, 10, 10};

  // mutual capacity for all commodities
  double mut_cap[numArcs] = {15, 15, 15, 15, 15};
  
  MatrixBlockIP N; // declare N (block constraints matrix) as a Matrix for BlockIP 
  int *srcN, *dstN;
  N.create_network_matrix(numArcs, numNodes, srcN, dstN); // N is created as network matrix
  // fill srcN and dstN; srcN and dstN allocated by create_network_matrix()
  for (int i = 0; i < numArcs; ++i) srcN[i] = src[i];
  for (int i = 0; i < numArcs; ++i) dstN[i] = dst[i];
    
  MatrixBlockIP L; // declare L (linking constraints matrix) as a Matrix for BlockIP
  L.create_identity_matrix(numArcs); // N is created as an identity  matrix
  
  BlockIP bip; // declare BlockIP problem
    
  double *cost, *qcost, *ub, *rhs;
  // creation of BlockIP problem
  bip.create_problem(BlockIP::QUADRATIC, cost, qcost, NULL, NULL, ub, rhs, numBlocks, true, &N, true, &L);
  
  // fill arrays of data for problem in standard form, previously allocated by create_problem()
  // linear and quadratic costs and bounds of block variables (multicommodity flows in this example)
  for (int i = 0; i < numBlocks*numArcs; ++i) {
    cost[i] = c[i];
    qcost[i] = 2*q[i]; // times 2 since BlockIP considers 1/2*x'*Q*x
    ub[i] = ind_cap[i];
  }
  // costs and bounds of slacks of linking constraints
  for (int i = 0; i < numArcs; ++i) {
    cost[numBlocks*numArcs + i] = 0;
    qcost[numBlocks*numArcs + i] = 0;
    ub[numBlocks*numArcs + i] = mut_cap[i];
  }
    
  // rhs of constraints
  // since block matrix was created as "network matrix", last row removed, then only numNodes-1 positions per block
  int ct = 0;
  for (int k = 0; k < numBlocks; ++k) {
    for (int i = 0; i < numNodes-1; ++i)
      rhs[ct++] = injection[k*numNodes + i];
  }
  for (int i = 0; i < numArcs; ++i)
    rhs[ct++] = mut_cap[i];

  // set some options
  bip.set_show_specrad(true); // compute and show estimation of spectral radius
  //solve
  cout << "Solve problem using PCG+Cholesky and Newton direction" << endl;
  bip.minimize(); // solve problem
  cout << "Total CGit= " << bip.get_cgit() << endl;
  
  // write problem in BlockIPformat, read problem (freed and created) and solve again with other options
  cout << endl << "Write BlockIP format" << endl;
  bip.write_BlockIP_format("example");
  cout << "Read  BlockIP format" << endl;
  bip.read_BlockIP_format("example.dat");
  // set some options
  bip.set_type_comp_dy(BlockIP::FULL_CHOL); // solve normal equation with Full Cholesky (not PCG+Cholesky)
  //solve
  cout << "Solve problem again using Cholesky and Newton direction" << endl;
  bip.minimize();

  // write problem in extended MPS format, read problem (freed and created) and solve again with other options
  cout << endl << "Write block-MPS format" << endl;
  bip.write_mps("example");
  cout << "Read  block-MPS format" << endl;
  bip.read_mps("example.mps");
  // set some options
  bip.set_type_direction(BlockIP::SECOND_ORDER); // computes predictor-corrector direction with PCG+Cholesky
  //solve
  cout << "Solve problem again using PCG+Cholesky and predictor-corrector direction" << endl;
  bip.minimize();
  cout << "Total CGit= " << bip.get_cgit() << endl;
  
  // solution as a nonlinear problem
  Param_fobj params; // parameters to compute the objective function
  params.nBlocks= numBlocks;
  params.nNodes= numNodes;
  params.nArcs= numArcs;
  params.c= &(c[0]);
  params.q= &(q[0]);
  
  // cost and qcost passed but not used, problem is nonlinear
  double *costX, *qcostX, *ubX, *rhsX;

  // N and L will be reused, then we have to free data of possible previous Cholesky factorizations
  // otherwise the new bipX problem will alloc new space and the previous one will never be freed
  // (causing a memory leak)
  N.chol.reset(); 
  L.chol.reset(); 
  BlockIP bipX(BlockIP::NONLINEAR, costX, qcostX, nonlin_objective_function, (void *) &params, ubX, rhsX, numBlocks, true, &N, true, &L);

  // another option would have been to use new NX and LX objects created 
  // with the copy constructor as 
  // MatrixBlockIP NX(&N); 
  // MatrixBlockIP LX(&L); 
  // BlockIP bipX(BlockIP::NONLINEAR, costX, qcostX, nonlin_objective_function, (void *) &params, ubX, rhsX, numBlocks, true, &NX, true, &LX);

  // ubX and rhsX computed as in previous bip quadratic instance
  for (int i = 0; i < numBlocks*numArcs; ++i)  ubX[i] = ind_cap[i];
  for (int i = 0; i < numArcs; ++i) ubX[numBlocks*numArcs + i] = mut_cap[i];
  ct = 0;
  for (int k = 0; k < numBlocks; ++k)
    for (int i = 0; i < numNodes-1; ++i) rhsX[ct++] = injection[k*numNodes + i];
  for (int i = 0; i < numArcs; ++i) rhsX[ct++] = mut_cap[i];
 
  // set some options
  bipX.set_show_specrad(true); // compute and show estimation of spectral radius
  //solve
  cout << endl << "Solve as NONLINEAR problem using PCG+Cholesky and Newton direction" << endl;
  bipX.minimize(); // solve problem
  cout << "Total CGit= " << bipX.get_cgit() << endl;

  return 0;

}
