/*!
  @file StdForm.h 
  @brief Definition of StdForm
*/

#ifndef STDFORM_H
#define STDFORM_H

#include <vector>

using namespace std;

//! Class for perform conversions between standard form and original problem
class StdForm {
  public:
    enum TYPE_TRANSFORM {NON_ZERO_LB, _INF_LB, FREE_VAR, X_, SLACK, SURPLUS};
    
    struct Transform { //!< To save data related to one transformation
      TYPE_TRANSFORM type; //!< Type of transformation
      double bound; //!< The bound when the type of transformation is Non Zero Lower Bound
                    //!< (x = x~ + lb) or -Infinity Lower Bound transformations (x = -x~ + ub)
      int x_index; //!< Index to the x- variable when the type of transformation is Free Variable
                   //!< (x = x+ - x-)
    };
    
    int numOrigVars; //!< Number of variables in the original problem
    int numTransfVars; //!< Number of variables in the transformed (standardized) problem
    int m_cons; //!< Number of constraints (including slacks) in the standardized problem

    int numTransforms; //!< Number of transformations performed in variables
    int *origVars; //!< Index to the original variables (that need a transformatiBon in objective function) in original problem, of size numTransforms
    int *transfVars; //!< Index to the original variables (that need a transformation in objective function) in standardized problem, of size numTransforms
    Transform *transforms; //!< Transformations performed in variables, of size numTransforms

    double constantFObj; //!< Constant to add in the objective function
    vector<int> *deletedRows; //!< Index to the rows that have been deleted
    
    double *xOrig; //!< Optimal solution in the original problem
    double *yOrig; //!< Optimal dual variables in the original problem
    double *GxOrig; //!< Gradient in x in the original problem
    double *HxOrig; //!< Hessian in x in the original problem
    double *wOrig; //!< Optimal dual variables of x_i <= u_i bounds in the original problem
    double *zOrig; //!< Optimal dual variables of x_i >= 0 bounds in the original problem
    
    
    //! User function to calculate the objective function in a point
    void (*fobj) (int n, double x[], double &fx, double Gx[], double Hx[], void *params);
    
    // Constructor
    StdForm();
    
    // Destructor
    ~StdForm();
    
    // Function to calculate the objective function in a transformed variables point
    void fobj_stdform(int n, double x[], double &fx, double Gx[], double Hx[], void *params);
    
    // Converts a point from transformed variables to original variables space
    void transformed_to_original_primal_variables(double* x);
    
    // Converts the dual variables of contraints from transformed to original variables space
    void transformed_to_original_dual_variables(double* y, double* w, double* z);
    
    // Converts a point from original variables to transformed variables space
    void original_to_transformed_primal_variables(double* &xout, double* xin);
    
    // Transforms objective function from original variables to transformed variables space
    void transform_linear_and_quadratic_cost(double* &cost, double* &qcost, double &constant);
    
    // Converts the variable names from original variables to transformed variables space
    void original_to_transformed_names(string* &outVarNames, string* inVarNames, string* &outConsNames, string* inConsNames);
    
    // Deletes the positions belonging to new variables
    void delete_new_variables(double* &xout, double* xin);
};

#endif //STDFORM_H
