#include "BlockIPInterface.h"
#include <iostream>
#include <fstream>
#include <time.h>
#include <cstring>
#include <math.h>

using namespace std; 

BlockIPInterface::BlockIPInterface()
//! Constructor
{
  bip = NULL;
  smlbip = NULL;
#ifdef OSI_INTERFACE
  oi = NULL;
#endif // OSI_INTERFACE
  solver = SOLVER::NONE;
  maxTime = -1;
  numThreads = -1;
  solved = false;
  logFilename = NULL;
  solFilename = NULL;
  stdForm = NULL;
  varNames = NULL;
  numVars = 0;
  varsOrder = NULL;
  type_input = TYPE_INPUT::NO_INPUT;
  x = NULL;
  y = NULL;
  z = NULL;
  w = NULL;
  infMatlab = 1.0e+32;
  iterMatlab = 200;
  typeCompMatlab = 1;
  pathMatlab = NULL;
  consNames = NULL;
  blockNames = NULL;
  varsPerBlock = NULL;
  consPerBlock = NULL;
  rc = NULL;
  indexUbInf = NULL;
}

BlockIPInterface::~BlockIPInterface()    
//! Destructor
{
  freeMemory();
}

void BlockIPInterface::setLogFilename(char *filename)
//! Set the name of the log file
/*!
  @note The filename will be copied
*/
{
  if (logFilename)
    delete[] logFilename;
  logFilename = new char[strlen(filename)+1];
  strcpy(logFilename, filename);
}

void BlockIPInterface::setSolFilename(char *filename)
//! Set the name of the solution file
/*!
  @note The filename will be copied
*/
{
  if (solFilename)
    delete[] solFilename;
  solFilename = new char[strlen(filename)+1];
  strcpy(solFilename, filename);
}

void BlockIPInterface::readMps(const char *filename)
//! Load a problem from a mps file
{
  if (solver == NONE)
    throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::SOLVER_NOT_DEFINED);
  if (smlbip) {
    delete smlbip; smlbip = NULL; bip = NULL;
  }
  else {
    delete bip; bip = NULL;
  }
#ifdef OSI_INTERFACE
  delete oi; oi = NULL;
#endif // OSI_INTERFACE
  delete stdForm; stdForm = NULL;
  delete[] varNames; varNames = NULL;
  delete[] varsOrder; varsOrder = NULL;
  delete[] x; x = NULL;
  delete[] y; y = NULL;
  delete[] z; z = NULL;
  delete[] w; w = NULL;
  delete[] consNames; consNames = NULL;
  delete[] blockNames; blockNames = NULL;
  delete[] varsPerBlock; varsPerBlock = NULL;
  delete[] consPerBlock; consPerBlock = NULL;
  solved = false;
  if (solver == BLOCKIP) {
    bip = new BlockIP();
    bip->read_mps(filename);
    varNames = bip->get_full_var_names();
    consNames = bip->get_full_cons_names();
  }
  else if (solver == BLOCKIP_MATLAB) {
    bip = new BlockIP();
    bip->read_mps(filename);
    numBlocks = bip->k_blocks;
    varsPerBlock = new int[numBlocks];
    consPerBlock = new int[numBlocks];
    for (int k = 0; k < numBlocks; ++k) {
      varsPerBlock[k] = bip->N[k].n;
      consPerBlock[k] = bip->N[k].m;
    }
    numLinCons = bip->l_link;
    numVars = bip->get_num_vars();
    varNames = bip->get_full_var_names();
    numCons = bip->get_num_cons();
    consNames = bip->get_full_cons_names();
    const string *auxBlockNames = NULL, *dummy;
    bip->get_names(auxBlockNames, dummy, dummy);
    blockNames = new string[numBlocks];
    for (int k = 0; k < numBlocks; ++k)
      blockNames[k] = auxBlockNames[k];
    bip->convert_to_standard();
  }
#ifdef OSI_INTERFACE      
  else {
    oi = new OsiInterface((OsiInterface::SOLVER) solver);
    oi->readMps(filename);
  }
#endif // OSI_INTERFACE  
  type_input = TYPE_INPUT::MPS;
}

void BlockIPInterface::readAmpl(const char *modelFilename, const char *dataFilename)
//! Load a problem from ampl files
{
  if (solver == NONE)
    throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::SOLVER_NOT_DEFINED);
  if (smlbip) {
    delete smlbip; smlbip = NULL; bip = NULL;
  }
  else {
    delete bip; bip = NULL;
  }
#ifdef OSI_INTERFACE
  delete oi; oi = NULL;
#endif // OSI_INTERFACE
  delete stdForm; stdForm = NULL;
  delete[] varNames; varNames = NULL;
  delete[] varsOrder; varsOrder = NULL;
  delete[] x; x = NULL;
  delete[] y; y = NULL;
  delete[] z; z = NULL;
  delete[] w; w = NULL;
  delete[] consNames; consNames = NULL;
  delete[] blockNames; blockNames = NULL;
  delete[] varsPerBlock; varsPerBlock = NULL;
  delete[] consPerBlock; consPerBlock = NULL;
  solved = false;
  smlbip = new SMLBlockIP();
  smlbip->readAmpl(modelFilename, dataFilename);
  bip = smlbip->getBlockIPInterface();
  if (solver != BLOCKIP && smlbip->isNonLinear())
    throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::NONLINEAR_NON_ACCEPTED);
  
  const int *auxVarsOrder = smlbip->getVarsOrder();
  numVars = bip->get_num_vars();
  varNames = new string[numVars];
  varsOrder = new int[numVars];
  string *unorderedVarNames = bip->get_full_var_names(); // ampl change the variables order
  for (int i = 0; i < numVars; ++i) {
    varsOrder[i] = auxVarsOrder[i];
    varNames[i] = unorderedVarNames[varsOrder[i]];
  }
  delete[] unorderedVarNames;
  
  
  if(solver == BLOCKIP) {
    // cout << "Using BlockIP, not implemented" << endl;
    consNames = bip->get_full_cons_names();
  }
  else if (solver == BLOCKIP_MATLAB) {
    numBlocks = bip->k_blocks;
    varsPerBlock = new int[numBlocks];
    consPerBlock = new int[numBlocks];
    for (int k = 0; k < numBlocks; ++k) {
      varsPerBlock[k] = bip->N[k].n;
      consPerBlock[k] = bip->N[k].m;
    }
    numLinCons = bip->l_link;
    numCons = bip->get_num_cons();
    consNames = bip->get_full_cons_names();
    const string *auxBlockNames = NULL, *dummy;
    bip->get_names(auxBlockNames, dummy, dummy);
    blockNames = new string[numBlocks];
    for (int k = 0; k < numBlocks; ++k)
      blockNames[k] = auxBlockNames[k];
    bip->convert_to_standard();
  }
#ifdef OSI_INTERFACE
  else {
    MatrixBlockIP *mbip = NULL;
    try {
      //to test stdForm
      // bip->convert_to_standard();
      mbip = new MatrixBlockIP();
      mbip->compute_full_matrix(bip->k_blocks, bip->sameN, bip->N, bip->sameL, bip->L);
    
      oi = new OsiInterface((OsiInterface::SOLVER) solver);
      if (bip->type_objective == BlockIP::LINEAR)
        oi->loadProblem(bip->m_cons, bip->n_vars, mbip->nz, bip->cost, NULL, bip->lb, bip->ub, bip->lhs, bip->rhs, mbip->inirowa, mbip->icola, mbip->a, true);
      else if (bip->type_objective == BlockIP::QUADRATIC)
        oi->loadProblem(bip->m_cons, bip->n_vars, mbip->nz, bip->cost, bip->qcost, bip->lb, bip->ub, bip->lhs, bip->rhs, mbip->inirowa, mbip->icola, mbip->a, true);
      
      // to test stdForm
      // oi->loadProblem(bip->m_cons, bip->n_vars, mbip->nz, bip->cost, NULL, bip->ub, bip->rhs, bip->rhs, mbip->inirowa, mbip->icola, mbip->a, true);
    } catch (exception &e) {
      delete mbip;
      throw;
    }
    delete mbip;
    delete smlbip; smlbip = NULL; bip = NULL; // delete smlbip also deletes bip
  }
#endif // OSI_INTERFACE  
  type_input = TYPE_INPUT::AMPL;
}

void BlockIPInterface::readBlockIPFormat(const char *filename)
//! Load a problem from BlockIP format file
{
  if (solver == NONE)
    throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::SOLVER_NOT_DEFINED);
  if (smlbip) {
    delete smlbip; smlbip = NULL; bip = NULL;
  }
  else {
    delete bip; bip = NULL;
  }
#ifdef OSI_INTERFACE
  delete oi; oi = NULL;
#endif // OSI_INTERFACE
  delete stdForm; stdForm = NULL;
  delete[] varNames; varNames = NULL;
  delete[] varsOrder; varsOrder = NULL;
  delete[] x; x = NULL;
  delete[] y; y = NULL;
  delete[] z; z = NULL;
  delete[] w; w = NULL;
  delete[] consNames; consNames = NULL;
  delete[] blockNames; blockNames = NULL;
  delete[] varsPerBlock; varsPerBlock = NULL;
  delete[] consPerBlock; consPerBlock = NULL;
  solved = false;
  bip = new BlockIP();
  bip->read_BlockIP_format(filename);
  if (solver != BLOCKIP && bip->type_objective == BlockIP::NONLINEAR)
    throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::NONLINEAR_NON_ACCEPTED);
  
  if(solver == BLOCKIP) {
    varNames = bip->get_full_var_names();
    consNames = bip->get_full_cons_names();
    // Nothing to do
  }
  else if (solver == BLOCKIP_MATLAB) {
    numBlocks = bip->k_blocks;
    varsPerBlock = new int[numBlocks];
    consPerBlock = new int[numBlocks];
    for (int k = 0; k < numBlocks; ++k) {
      int index = bip->sameN ? 0 : k;
      varsPerBlock[k] = bip->N[index].n;
      consPerBlock[k] = bip->N[index].m;
    }
    numLinCons = bip->l_link;
    numVars = bip->get_num_vars();
    varNames = bip->get_full_var_names();
    numCons = bip->get_num_cons();
    consNames = bip->get_full_cons_names();
    const string *auxBlockNames = NULL, *dummy;
    bip->get_names(auxBlockNames, dummy, dummy);
    blockNames = new string[numBlocks];
    for (int k = 0; k < numBlocks; ++k)
      blockNames[k] = auxBlockNames[k];
    bip->convert_to_standard();
  }
#ifdef OSI_INTERFACE
  else {
    MatrixBlockIP *mbip = NULL;
    try {
      numVars = bip->get_num_vars();
      varNames = bip->get_full_var_names();
      //to test stdForm
      // bip->convert_to_standard();
      mbip = new MatrixBlockIP();
      mbip->compute_full_matrix(bip->k_blocks, bip->sameN, bip->N, bip->sameL, bip->L);

      oi = new OsiInterface((OsiInterface::SOLVER) solver);
      if (bip->inStdForm) {
        if (bip->type_objective == BlockIP::LINEAR)
          oi->loadProblem(bip->m_cons, bip->n_vars, mbip->nz, bip->cost, NULL, NULL, bip->ub, bip->rhs, bip->rhs, mbip->inirowa, mbip->icola, mbip->a, true);
        else if (bip->type_objective == BlockIP::QUADRATIC)
          oi->loadProblem(bip->m_cons, bip->n_vars, mbip->nz, bip->cost, bip->qcost, NULL, bip->ub, bip->rhs, bip->rhs, mbip->inirowa, mbip->icola, mbip->a, true);
      }
      else {
        if (bip->type_objective == BlockIP::LINEAR)
          oi->loadProblem(bip->m_cons, bip->n_vars, mbip->nz, bip->cost, NULL, bip->lb, bip->ub, bip->lhs, bip->rhs, mbip->inirowa, mbip->icola, mbip->a, true);
        else if (bip->type_objective == BlockIP::QUADRATIC)
          oi->loadProblem(bip->m_cons, bip->n_vars, mbip->nz, bip->cost, bip->qcost, bip->lb, bip->ub, bip->lhs, bip->rhs, mbip->inirowa, mbip->icola, mbip->a, true);
      }
      
      // to test stdForm
      // oi->loadProblem(bip->m_cons, bip->n_vars, mbip->nz, bip->cost, NULL, bip->ub, bip->rhs, bip->rhs, mbip->inirowa, mbip->icola, mbip->a, true);
    } catch (exception &e) {
      delete mbip;
      throw;
    }
    delete mbip;
    delete bip; bip = NULL;
  }
#endif // OSI_INTERFACE  
  type_input = TYPE_INPUT::BLOCKIP_FORMAT;
}

void BlockIPInterface::createMatlabProblem()
//! Create the files to load and solve a problem in BlockIP Matlab
{
  ofstream outFile ("tmpBlockIP.m", ios::out | ios::trunc);
  if (outFile.is_open()) {
    numVarsMatlab = bip->n_vars;
    numConsMatlab = bip->m_cons;
    int n_vars = bip->n_vars;
    
    outFile << "c = [" << endl;
    for (int i = 0; i < n_vars; ++i)
      outFile << bip->cost[i] << endl;
    outFile << "];\n" << endl;
    
    outFile << "q = [" << endl;
    if (bip->qcost) {
      for (int i = 0; i < n_vars; ++i)
        outFile << bip->qcost[i] << endl;
    }
    outFile << "];\n" << endl;
    
    delete indexUbInf;
    indexUbInf = new vector<int>;
    outFile << "u = [" << endl;
    double inf = bip->get_inf();
    double minim = 1;
    for (int i = 0; i < n_vars; ++i)
      if (bip->ub[i] < minim && bip->ub[i] != 0)
        minim = bip->ub[i];
    double zero = min(1.0e-6, minim/1.0e-6);
      
    for (int i = 0; i < n_vars; ++i) {
      if (bip->ub[i] >= inf) {
        indexUbInf->push_back(i);
        outFile << infMatlab << endl;
      }
      else if (bip->ub[i] == 0)
        outFile << zero << endl;
      else
        outFile << bip->ub[i] << endl;
    }
    outFile << "];\n" << endl;
    
    outFile << "b = [" << endl;
    for (int i = 0; i < bip->m_cons; ++i)
      outFile << bip->rhs[i] << endl;
    outFile << "];\n" << endl;
    
    int nblocks = bip->sameN ? 1 : bip->k_blocks;
    if (!bip->sameN)
      outFile << "N = cell (" << nblocks << ",1);" << endl;
    for (int k = 0; k < nblocks; ++k) {
      outFile << "Nindex" << k+1 << " = [" << endl;
      if (bip->N[k].columnwise) {
        for (int j = 0; j < bip->N[k].n; ++j)
          for (int i = bip->N[k].inicola[j]; i < bip->N[k].inicola[j+1]; ++i)
            outFile << bip->N[k].irowa[i]+1 << " " << j+1 << " " << bip->N[k].a[i] << endl;
      }
      else {
        if (!bip->N[k].rowwise)
          bip->N[k].native_to_general();
        for (int i = 0; i < bip->N[k].m; ++i)
          for (int j = bip->N[k].inirowa[i]; j < bip->N[k].inirowa[i+1]; ++j)
            outFile << i+1 << " " << bip->N[k].icola[j]+1 << " " << bip->N[k].a[j] << endl;
      }
      outFile << "];" << endl;
      if (!bip->sameN)
        outFile << "N{" << k+1 << "} = sparse(Nindex" << k+1 << "(:,1), Nindex" << k+1 <<
        "(:,2), Nindex" << k+1 << "(:,3), " << bip->N[k].m << ", " << bip->N[k].n << ");\n" << endl;
      else
        outFile << "N = sparse(Nindex" << k+1 << "(:,1), Nindex" << k+1 <<
        "(:,2), Nindex" << k+1 << "(:,3), " << bip->N[k].m << ", " << bip->N[k].n << ");\n" << endl;
    }
    
    int lblocks = bip->sameL ? 1 : bip->k_blocks;
    if (!bip->sameL)
      outFile << "L = cell (" << lblocks << ",1);" << endl;
    for (int k = 0; k < lblocks; ++k) {
      outFile << "Lindex" << k+1 << " = [" << endl;
      if (bip->L[k].columnwise) {
        for (int j = 0; j < bip->L[k].n; ++j)
          for (int i = bip->L[k].inicola[j]; i < bip->L[k].inicola[j+1]; ++i)
            outFile << bip->L[k].irowa[i]+1 << " " << j+1 << " " << bip->L[k].a[i] << endl;
      }
      else {
        if (!bip->L[k].rowwise)
          bip->L[k].native_to_general();
        for (int i = 0; i < bip->L[k].m; ++i)
          for (int j = bip->L[k].inirowa[i]; j < bip->L[k].inirowa[i+1]; ++j)
            outFile << i+1 << " " << bip->L[k].icola[j]+1 << " " << bip->L[k].a[j] << endl;
      }
      outFile << "];" << endl;
      if (!bip->sameL)
        outFile << "L{" << k+1 << "} = sparse(Lindex" << k+1 << "(:,1), Lindex" << k+1 <<
        "(:,2), Lindex" << k+1 << "(:,3), " << bip->L[k].m << ", " << bip->L[k].n << ");\n" << endl;
      else
        outFile << "L = sparse(Lindex" << k+1 << "(:,1), Lindex" << k+1 <<
        "(:,2), Lindex" << k+1 << "(:,3), " << bip->L[k].m << ", " << bip->L[k].n << ");\n" << endl;
    }

    if (pathMatlab) {
      outFile << "path(path,'" << pathMatlab << "');" << endl;    
      outFile << "path(path,'" << pathMatlab << "/ng-peyton');" << endl;
      outFile << "path(path,'" << pathMatlab << "/split_mex');" << endl;
      outFile << "path(path,'" << pathMatlab << "/valRitz_mex');" << endl;
    }
    else {
      outFile << "path(path,'.');" << endl;    
      outFile << "path(path,'./ng-peyton');" << endl;
      outFile << "path(path,'./split_mex');" << endl;
      outFile << "path(path,'./valRitz_mex');" << endl;
    }
    outFile << "[status,x,y,z,w,pobj,dobj,activelnk,it,cgit] = "
    "prblock_ip(c,q,[],u,b,N,L,[]," << infMatlab << "," << iterMatlab << 
    "," << typeCompMatlab << ",'sprsblkllt',0,1e-6);\n" << endl;
    outFile << "dlmwrite('stBlockIP.tmp', status, ' ');" << endl;
    outFile << "dlmwrite('xBlockIP.tmp', x, ' ');" << endl;
    outFile << "dlmwrite('objBlockIP.tmp', pobj, ' ');" << endl;
    outFile << "dlmwrite('dobjBlockIP.tmp', dobj, ' ');" << endl;
    outFile << "dlmwrite('itBlockIP.tmp', it, ' ');" << endl;
    outFile << "dlmwrite('cgitBlockIP.tmp', cgit, ' ');" << endl;
    outFile << "dlmwrite('yBlockIP.tmp', y, ' ');" << endl;
    outFile << "dlmwrite('zBlockIP.tmp', z, ' ');" << endl;
    outFile << "dlmwrite('wBlockIP.tmp', w, ' ');" << endl;
    outFile << "quit;" << endl;
    outFile.close();
  }
  else 
    throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::CREATE_TMP_FILES);
#ifdef WIN32
  outFile.open ("tmpExecMatlab.bat", ios::out | ios::trunc);
#else
  outFile.open ("tmpExecMatlab", ios::out | ios::trunc);  
#endif
  if (outFile.is_open()) {
    if (logFilename)
      outFile << "matlab -nodisplay -nojvm -nodesktop -nosplash "
      "-r tmpBlockIP.m >& " << logFilename << endl;
    else
      outFile << "matlab -nodisplay -nojvm -nodesktop -nosplash "
      "-r tmpBlockIP.m" << endl;
      // outFile << "nohup matlab -nodisplay -nojvm -nodesktop -nosplash "
      // "-r tmpBlockIP.m" << endl;
    outFile.close();
    #ifndef WIN32
      system("chmod 700 tmpExecMatlab");
    #endif
  }
  else
    throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::CREATE_TMP_FILES);
}

void BlockIPInterface::amplToMps(const char *modelFilename, const char *dataFilename, const char *mpsFilename, bool convertToStd)
//! Converts a problem in ampl format to mps format
/*!
  @param convertToStd If true the problem will be written in BlockIP standard form
*/
{
  smlbip->amplToMps(modelFilename, dataFilename, mpsFilename, convertToStd);
}

void BlockIPInterface::amplToBlockIPFormat(const char *modelFilename, const char *dataFilename, const char *BlockIPFormatFilename, bool convertToStd)
//! Converts a problem in ampl format to BlockIP format
/*!
  @param convertToStd If true the problem will be written in BlockIP standard form
*/
{
  smlbip->amplToBlockIPFormat(modelFilename, dataFilename, BlockIPFormatFilename, convertToStd);
}

void BlockIPInterface::mpsToBlockIPFormat(const char *mpsFilename, const char *BlockIPFormatFilename, bool convertToStd)
//! Converts a problem in mps format to BlockIP format
/*!
  @param convertToStd If true the problem will be written in BlockIP standard form
*/
{
  BlockIP *bip = new BlockIP();
  try {
    bip->read_mps(mpsFilename);
    if (convertToStd)
      bip->convert_to_standard();
    bip->write_BlockIP_format(BlockIPFormatFilename);
  } catch (exception &e) {
    delete bip;
    throw;
  }
  delete bip;
}

void BlockIPInterface::BlockIPFormatToMps(const char *BlockIPFormatFilename, const char *mpsFilename, bool convertToStd)
//! Converts a problem in BlockIP format to mps format
/*!
  @param convertToStd If true the problem will be written in BlockIP standard form
*/
{
  BlockIP *bip = new BlockIP();
  try {
    bip->read_BlockIP_format(BlockIPFormatFilename);
    if (convertToStd)
      bip->convert_to_std_and_write_mps(mpsFilename);
    else
      bip->write_mps(mpsFilename);
  } catch (exception &e) {
    delete bip;
    throw;
  }
  delete bip;
}


BlockIPInterface::OPT_STATUS BlockIPInterface::solve()
//! Solve the problem loaded
{
  OPT_STATUS ret = OTHERWISE;
  bool solutionFound = false;
  clock_t t, t0;
  t0= clock();
  if (solver == BLOCKIP) {  
    // bip->set_max_time(maxTime);
    // bip->set_num_threads(numThreads);
    // bip->set_log_filename(logFilename);
    // ret = bip->solve();
    switch (bip->minimize()) {
      case 0:
        ret = OPTIMAL_SOLUTION;
        solutionFound = true;
        break;
      case 1:
        ret = INFEASIBLE;
        break;
    }
  }
  else if (solver == BLOCKIP_MATLAB) {
    createMatlabProblem();
    stdForm = bip->get_std_form(true);
    if (smlbip) {
      delete smlbip; smlbip = NULL; bip = NULL; // delete smlbip also deletes bip
    }
    else {
      delete bip; bip = NULL;
    }
#ifdef WIN32
    system("tmpExecMatlab.bat");
#else
    system("tmpExecMatlab < /dev/null");
#endif
    ifstream inFile("stBlockIP.tmp", ios::in);
    if (!inFile.is_open())
      throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::OPEN_TMP_FILES);
    int st;
    if (inFile >> st) {
      switch (st) {
        case 0:
          ret = OPTIMAL_SOLUTION;
          solutionFound = true;
          break;
        case 1:
          ret = INFEASIBLE;
          break;
        case 2:
          ret = UNBOUNDED;
          solutionFound = true;
          break;
      }
    }
    inFile.close();
    if (ret == OPTIMAL_SOLUTION || ret == UNBOUNDED) {
      inFile.open("objBlockIP.tmp", ios::in);
      if (!inFile.is_open())
        throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::OPEN_TMP_FILES);
        
      inFile >> fobj;
      inFile.close();
      
      inFile.open("dobjBlockIP.tmp", ios::in);
      if (!inFile.is_open())
        throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::OPEN_TMP_FILES);
        
      inFile >> dobj;
      inFile.close();
      inFile.open("itBlockIP.tmp", ios::in);
      if (!inFile.is_open())
        throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::OPEN_TMP_FILES);
        
      inFile >> iter;
      inFile.close();
      inFile.open("cgitBlockIP.tmp", ios::in);
      if (!inFile.is_open())
        throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::OPEN_TMP_FILES);
        
      inFile >> iterPCG;
      inFile.close();
      inFile.open("xBlockIP.tmp", ios::in);
      if (!inFile.is_open())
        throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::OPEN_TMP_FILES);
        
      x = new double[numVarsMatlab];
      for (int i = 0; i < numVarsMatlab; ++i)
        inFile >> x[i];
        
      inFile.close();
      inFile.open("yBlockIP.tmp", ios::in);
      if (!inFile.is_open())
        throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::OPEN_TMP_FILES);
        
      y = new double[numConsMatlab];
      for (int i = 0; i < numConsMatlab; ++i)
        inFile >> y[i];
        
      inFile.close();
      inFile.open("zBlockIP.tmp", ios::in);
      if (!inFile.is_open())
        throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::OPEN_TMP_FILES);
        
      try {
        z = new double[numVarsMatlab];
        for (int i = 0; i < numVarsMatlab; ++i)
          inFile >> z[i];
          
        inFile.close();
        inFile.open("wBlockIP.tmp", ios::in);
        if (!inFile.is_open())
        throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::OPEN_TMP_FILES);
        
        w = new double[numVarsMatlab];
        if (!indexUbInf->size())
          for (int i = 0; i < numVarsMatlab; ++i)
            inFile >> w[i];
        else {
          int j = 0;
          for (int i = 0; i < numVarsMatlab; ++i) {
            if (j < indexUbInf->size() && (*indexUbInf)[j] == i) {
              w[i] = NAN;
              ++j;
            }
            else
              inFile >> w[i];
          }
        }
        delete indexUbInf; indexUbInf = NULL;
      } catch (exception &e) {
        delete[] z; delete[] w;
        throw;
      }
    }
    else
      ret = INFEASIBLE;
    inFile.close();
#ifdef WIN32
    system("del tmpBlockIP.m tmpExecMatlab.bat xBlockIP.tmp objBlockIP.tmp "
    "dobjBlockIP.tmp itBlockIP.tmp cgitBlockIP.tmp yBlockIP.tmp zBlockIP.tmp "
    "wBlockIP.tmp stBlockIP.tmp");
#else
    system("/bin/rm tmpBlockIP.m tmpExecMatlab xBlockIP.tmp objBlockIP.tmp "
    "dobjBlockIP.tmp itBlockIP.tmp cgitBlockIP.tmp yBlockIP.tmp zBlockIP.tmp "
    "wBlockIP.tmp stBlockIP.tmp");
#endif
  }
#ifdef OSI_INTERFACE
  else {
    oi->setMaxTime(maxTime);
    if (numThreads != -1)
      oi->setNumThreads(numThreads);
    oi->setLogFileName(logFilename);
    OsiInterface::OPT_STATUS status = oi->solve();
    switch (status) {
      case OPTIMAL_SOLUTION:
        ret = OPTIMAL_SOLUTION;
        solutionFound = true;
        break;
      case TIME_LIMIT_INFEAS:
        ret = TIME_LIMIT_INFEAS;
        break;
      case TIME_LIMIT_FEAS:
        ret = TIME_LIMIT_FEAS;
        solutionFound = true;
        break;
      case FEASIBLE:
        ret = FEASIBLE;
        solutionFound = true;
        break;
      case INFEASIBLE:
        ret = INFEASIBLE;
        break;
      default:
        ret = OTHERWISE;
    }
  }
#endif // OSI_INTERFACE
  t= clock();
  CPUsec= (double) (t-t0)/CLOCKS_PER_SEC;

  if (solutionFound) {
    const double *auxX = NULL;
    const double *auxY = NULL;
    const double *auxZ = NULL;
    const double *auxW = NULL;
    if (solver == BLOCKIP) {
      fobj = bip->get_fobj();
      dobj = bip->get_dobj();
      numVars = bip->get_num_vars();
      numCons = bip->get_num_cons();
      iter = bip->get_it();
      iterPCG = bip->get_cgit();
      auxX = bip->get_x();
      auxY = bip->get_y();
      auxZ = bip->get_z();
      auxW = bip->get_w();
      
      x = new double[numVars];
      y = new double[numCons];
      z = new double[numVars];
      w = new double[numVars];
      if (type_input == TYPE_INPUT::AMPL) { // order variables
        for (int i = 0; i < numVars; ++i) {
          x[i] = auxX[varsOrder[i]];
          z[i] = auxZ[varsOrder[i]];
          w[i] = auxW[varsOrder[i]];
        }
        for (int i = 0; i < numCons; ++i)
          y[i] = auxY[i];
      }
      else
        for (int i = 0; i < numVars; ++i) {
          x[i] = auxX[i];
          z[i] = auxZ[i];
          w[i] = auxW[i];
        }
        for (int i = 0; i < numCons; ++i)
          y[i] = auxY[i];
          
      rc = new double[numVars];
      for (int i = 0; i < numVars; ++i)
        rc[i] = (z[i] || isnan(w[i])) ? z[i] : w[i];
    }
    else if (solver != BLOCKIP_MATLAB) {
#ifdef OSI_INTERFACE
      fobj = oi->getObjValue();
      numVars = oi->getNumVars();
      auxX = oi->getSolution();
      
      x = new double[numVars];
      if (type_input == TYPE_INPUT::AMPL) { // order variables
        for (int i = 0; i < numVars; ++i)
          x[i] = auxX[varsOrder[i]];
      }
      else
        for (int i = 0; i < numVars; ++i)
          x[i] = auxX[i];
#endif // OSI_INTERFACE
    }
    else if (solver == BLOCKIP_MATLAB) {
      // convert from extended to original variables
      if (stdForm) {
        fobj += stdForm->constantFObj;
        dobj += stdForm->constantFObj;
        stdForm->transformed_to_original_primal_variables(x);
        auxX = stdForm->xOrig;
        delete[] x;
        stdForm->transformed_to_original_dual_variables(y, w, z);
        auxY = stdForm->yOrig;
        auxZ = stdForm->zOrig;
        auxW = stdForm->wOrig;
        delete[] y;
        delete[] w;
        delete[] z;
      }
      
      if (type_input == TYPE_INPUT::AMPL) { // order variables
        if (!stdForm) {
          auxX = x;
          auxY = y;
          auxZ = z;
          auxW = w;
        }
        x = new double[numVars];
        y = new double[numCons];
        z = new double[numVars];
        w = new double[numVars];
        for (int i = 0; i < numVars; ++i) {
          x[i] = auxX[varsOrder[i]];
          z[i] = auxZ[varsOrder[i]];
          w[i] = auxW[varsOrder[i]];
        }
        for (int i = 0; i < numCons; ++i)
          y[i] = auxY[i];
        
        if (!stdForm) {
          delete[] auxX;
          delete[] auxY;
          delete[] auxZ;
          delete[] auxW;
        }
      }
      else if (stdForm) {
        x = new double[numVars];
        y = new double[numCons];
        z = new double[numVars];
        w = new double[numVars];
        for (int i = 0; i < numVars; ++i) {
          x[i] = auxX[i];
          z[i] = auxZ[i];
          w[i] = auxW[i];
        }
        for (int i = 0; i < numCons; ++i)
          y[i] = auxY[i];
      }
      
      rc = new double[numVars];
      for (int i = 0; i < numVars; ++i)
        rc[i] = (z[i] || isnan(w[i])) ? z[i] : w[i];
    }
    
    if (solFilename) {
      ofstream outFile (solFilename, ios::out | ios::trunc);
      if (outFile.is_open()) {
        if (solver == BLOCKIP_MATLAB || solver == BLOCKIP) {
          for (int i = 0; i < numVars; ++i) {
            outFile << varNames[i] << "\t" << x[i] << "\t" << z[i] << "\t";
            if (isnan(w[i]))
              outFile << '.' << "\n";
            else
              outFile << w[i] << "\n";
          }
          outFile << "\n";
          for (int i = 0; i < numCons; ++i)
            outFile << consNames[i] << "\t" << y[i] << "\n";
        }
#ifdef OSI_INTERFACE
        else {
          if (type_input == TYPE_INPUT::MPS) {
            numVars = oi->getNumVars();
            const string *auxVarNames = oi->getVarNames();
            varNames = new string[numVars];
            for (int i = 0; i < numVars; ++i)
              varNames[i] = auxVarNames[i];
          }
          if (varNames)
            for (int i = 0; i < numVars; ++i)
              outFile << varNames[i] << "\t" << x[i] << "\n";
          else
            for (int i = 0; i < numVars; ++i)
              outFile << x[i] << "\n";
        }
#endif // OSI_INTERFACE
        outFile.close();
      }
      else
        throw ExceptionBlockIPInterface(ExceptionBlockIPInterface::SOLUTION_FILE);
    }
  }
  return ret;
}

int BlockIPInterface::getNumVars()
//! Get the number of variables
/*!
  @return Number of variables
*/
{
  if (bip && (solver == BLOCKIP || solver == BLOCKIP_MATLAB))
    return bip->get_num_vars();
  else {
#ifdef OSI_INTERFACE
    if (oi)
      return oi->getNumVars();
#endif // OSI_INTERFACE
    return 0;
  }
}

const double* BlockIPInterface::getSolution()
//! Get the solution
/*!
  @return Solution
*/
{
  return x;
}

void BlockIPInterface::freeMemory()
//! Free all allocated memory
{
  if (smlbip) {
    delete smlbip; smlbip = NULL; bip = NULL;
  }
  else {
    delete bip; bip = NULL;
  }
#ifdef OSI_INTERFACE
  delete oi; oi = NULL;
#endif // OSI_INTERFACE
  delete stdForm; stdForm = NULL;
  delete[] logFilename; logFilename = NULL;
  delete[] solFilename; solFilename = NULL;
  delete[] varNames; varNames = NULL;
  delete[] x; x = NULL;
  delete[] y; y = NULL;
  delete[] z; z = NULL;
  delete[] w; w = NULL;
  delete[] varsOrder; varsOrder = NULL;
  delete[] consNames; consNames = NULL;
  delete[] blockNames; blockNames = NULL;
  delete[] varsPerBlock; varsPerBlock = NULL;
  delete[] consPerBlock; consPerBlock = NULL;
  delete[] rc; rc = NULL;
  delete indexUbInf; indexUbInf = NULL;
}

// Display the primal variables value of blocks and slacks
void BlockIPInterface::printX()
//! Show all the primal variables of blocks and slacks
{
  for (int i = 0; i < numVars; ++i)
    cout << varNames[i] << " " << x[i] << endl;
}

void BlockIPInterface::printX(int globalVarIndex)
//! Show the value of the i-th primal variable
{
  if (globalVarIndex < 0 || globalVarIndex >= numVars)
    cerr << "Index out of range" << endl;
  else
    cout << varNames[globalVarIndex] << " " << x[globalVarIndex] << endl;
}

void BlockIPInterface::printX(string varName)
//! Show the value of the named primal variable
{
  bool writed = false;
  for (int i = 0; i < numVars; ++i)
    if (varName == varNames[i]) {
      cout << varNames[i] << " " << x[i] << endl;
      writed = true;
      break;
    }
  if (!writed)
    cerr << "There is no variable with that name" << endl;
}

void BlockIPInterface::printX(int blockIndex, int varIndex)
//! Show the value of the i-th primal variable inside the block k
{
  if (blockIndex < 0 || blockIndex >= numBlocks)
    cerr << "Block index out of range" << endl;
  else if (varIndex < 0 || varIndex >= varsPerBlock[blockIndex])
    cerr << "Variable index out of range" << endl;
  else {
    int index = 0;
    for (int i = 0; i < blockIndex; ++i)
      index += varsPerBlock[i];
    cout << varNames[index+varIndex] << " " << x[index+varIndex] << endl;
  }
}

void BlockIPInterface::printX(string blockName, int varIndex)
//! Show the value of the i-th primal variable inside the named block
{
  int k;
  int index = 0;
  for (k = 0; k < numBlocks; ++k) {
    if (blockName == blockNames[k]) {
      if (varIndex < 0 || varIndex >= varsPerBlock[k])
        cerr << "Variable index out of range" << endl;
      else
        cout << varNames[index+varIndex] << " " << x[index+varIndex] << endl;
      return;
    }
    index += varsPerBlock[k];
  }
  cerr << "There is no block with that name" << endl;
}

void BlockIPInterface::printXBlock(int blockIndex)
//! Show the value of all primal variables inside the block k
{
  if (blockIndex < 0 || blockIndex >= numBlocks)
    cerr << "Block index out of range" << endl;
  else {
    int index = 0;
    for (int i = 0; i < blockIndex; ++i)
      index += varsPerBlock[i];
    for (int i = 0; i < varsPerBlock[blockIndex]; ++i)
      cout << varNames[index+i] << " " << x[index+i] << endl;
  }
}

void BlockIPInterface::printXBlock(string blockName)
//! Show the value of all primal variables inside the named block
{
  int k;
  int index = 0;
  for (k = 0; k < numBlocks; ++k) {
    if (blockName == blockNames[k]) {
      for (int i = 0; i < varsPerBlock[k]; ++i)
        cout << varNames[index+i] << " " << x[index+i] << endl;
      return;
    }
    index += varsPerBlock[k];
  }
  cerr << "There is no block with that name" << endl;
}

void BlockIPInterface::printXSlack()
//! Show the value of all primal variables inside the slacks
{
  for (int i = numVars-numLinCons; i < numVars; ++i)
    cout << varNames[i] << " " << x[i] << endl;
}

void BlockIPInterface::printXSlack(int slackIndex)
//! Show the value of the i-th primal variable inside the slacks
{
  if (slackIndex < 0 || slackIndex >= numLinCons)
    cerr << "Index out of range" << endl;
  else
    cout << varNames[numVars-numLinCons+slackIndex] << " " << x[numVars-numLinCons+slackIndex] << endl;
}


// Display the reduced cost of primal variables of blocks and slacks
void BlockIPInterface::printRC()
//! Show all the reduced cost of primal variables of blocks and slacks
{
  for (int i = 0; i < numVars; ++i)
    cout << varNames[i] << " " << rc[i] << endl;
}

void BlockIPInterface::printRC(int globalVarIndex)
//! Show the reduced cost of the i-th primal variable
{
  if (globalVarIndex < 0 || globalVarIndex >= numVars)
    cerr << "Index out of range" << endl;
  else
    cout << varNames[globalVarIndex] << " " << rc[globalVarIndex] << endl;
}

void BlockIPInterface::printRC(string varName)
//! Show the reduced cost of the named primal variable
{
  bool writed = false;
  for (int i = 0; i < numVars; ++i)
    if (varName == varNames[i]) {
      cout << varNames[i] << " " << rc[i] << endl;
      writed = true;
      break;
    }
  if (!writed)
    cerr << "There is no variable with that name" << endl;
}

void BlockIPInterface::printRC(int blockIndex, int varIndex)
//! Show the reduced cost of the i-th primal variable inside the block k
{
  if (blockIndex < 0 || blockIndex >= numBlocks)
    cerr << "Block index out of range" << endl;
  else if (varIndex < 0 || varIndex >= varsPerBlock[blockIndex])
    cerr << "Variable index out of range" << endl;
  else {
    int index = 0;
    for (int i = 0; i < blockIndex; ++i)
      index += varsPerBlock[i];
    cout << varNames[index+varIndex] << " " << rc[index+varIndex] << endl;
  }
}

void BlockIPInterface::printRC(string blockName, int varIndex)
//! Show the reduced cost of the i-th primal variable inside the named block
{
  int k;
  int index = 0;
  for (k = 0; k < numBlocks; ++k) {
    if (blockName == blockNames[k]) {
      if (varIndex < 0 || varIndex >= varsPerBlock[k])
        cerr << "Variable index out of range" << endl;
      else
        cout << varNames[index+varIndex] << " " << rc[index+varIndex] << endl;
      return;
    }
    index += varsPerBlock[k];
  }
  cerr << "There is no block with that name" << endl;
}

void BlockIPInterface::printRCBlock(int blockIndex)
//! Show the reduced cost of all primal variables inside the block k
{
  if (blockIndex < 0 || blockIndex >= numBlocks)
    cerr << "Block index out of range" << endl;
  else {
    int index = 0;
    for (int i = 0; i < blockIndex; ++i)
      index += varsPerBlock[i];
    for (int i = 0; i < varsPerBlock[blockIndex]; ++i)
      cout << varNames[index+i] << " " << rc[index+i] << endl;
  }
}

void BlockIPInterface::printRCBlock(string blockName)
//! Show the reduced cost of all primal variables inside the named block
{
  int k;
  int index = 0;
  for (k = 0; k < numBlocks; ++k) {
    if (blockName == blockNames[k]) {
      for (int i = 0; i < varsPerBlock[k]; ++i)
        cout << varNames[index+i] << " " << rc[index+i] << endl;
      return;
    }
    index += varsPerBlock[k];
  }
  cerr << "There is no block with that name" << endl;
}

void BlockIPInterface::printRCSlack()
//! Show the reduced cost of all primal variables inside the slacks
{
  for (int i = numVars-numLinCons; i < numVars; ++i)
    cout << varNames[i] << " " << rc[i] << endl;
}

void BlockIPInterface::printRCSlack(int slackIndex)
//! Show the reduced cost of the i-th primal variable inside the slacks
{
  if (slackIndex < 0 || slackIndex >= numLinCons)
    cerr << "Index out of range" << endl;
  else
    cout << varNames[numVars-numLinCons+slackIndex] << " " << rc[numVars-numLinCons+slackIndex] << endl;
}


// Display the dual variables value of blocks and linking constraints
void BlockIPInterface::printCons()
//! Show all the dual variables of blocks and linking constraints
{
  for (int i = 0; i < numCons; ++i)
    cout << consNames[i] << " " << y[i] << endl;
}

void BlockIPInterface::printCons(int globalConsIndex)
//! Show the dual variable of the i-th constraint
{
  if (globalConsIndex < 0 || globalConsIndex >= numVars)
    cerr << "Index out of range" << endl;
  else
    cout << consNames[globalConsIndex] << " " << y[globalConsIndex] << endl;
}

void BlockIPInterface::printCons(string consName)
//! Show the dual variable of the named constraint
{
  bool writed = false;
  for (int i = 0; i < numCons; ++i)
    if (consName == consNames[i]) {
      cout << consNames[i] << " " << y[i] << endl;
      writed = true;
      break;
    }
  if (!writed)
    cerr << "There is no constraint with that name" << endl;
}

void BlockIPInterface::printCons(int blockIndex, int consIndex)
//! Show the dual variable of the i-th constraint inside the block k
{
  if (blockIndex < 0 || blockIndex >= numBlocks)
    cerr << "Block index out of range" << endl;
  else if (consIndex < 0 || consIndex >= consPerBlock[blockIndex])
    cerr << "Constraint index out of range" << endl;
  else {
    int index = 0;
    for (int i = 0; i < blockIndex; ++i)
      index += consPerBlock[i];
    cout << consNames[index+consIndex] << " " << y[index+consIndex] << endl;
  }
}

void BlockIPInterface::printCons(string blockName, int consIndex)
//! Show the dual variable of the i-th constraint inside the named block
{
  int k;
  int index = 0;
  for (k = 0; k < numBlocks; ++k) {
    if (blockName == blockNames[k]) {
      if (consIndex < 0 || consIndex >= consPerBlock[k])
        cerr << "Constraint index out of range" << endl;
      else
        cout << consNames[index+consIndex] << " " << y[index+consIndex] << endl;
      return;
    }
    index += consPerBlock[k];
  }
  cerr << "There is no block with that name" << endl;
}

void BlockIPInterface::printConsBlock(int blockIndex)
//! Show all the dual variables inside the block k
{
  if (blockIndex < 0 || blockIndex >= numBlocks)
    cerr << "Block index out of range" << endl;
  else {
    int index = 0;
    for (int i = 0; i < blockIndex; ++i)
      index += consPerBlock[i];
    for (int i = 0; i < consPerBlock[blockIndex]; ++i)
      cout << consNames[index+i] << " " << y[index+i] << endl;
  }
}

void BlockIPInterface::printConsBlock(string blockName)
//! Show all the dual varialbes inside the named block
{
  int k;
  int index = 0;
  for (k = 0; k < numBlocks; ++k) {
    if (blockName == blockNames[k]) {
      for (int i = 0; i < consPerBlock[k]; ++i)
        cout << consNames[index+i] << " " << y[index+i] << endl;
      return;
    }
    index += consPerBlock[k];
  }
  cerr << "There is no block with that name" << endl;
}

void BlockIPInterface::printConsLinking()
//! Show all the dual varialbes inside the linking constraints
{
  for (int i = numCons-numLinCons; i < numCons; ++i)
    cout << consNames[i] << " " << y[i] << endl;
}

void BlockIPInterface::printConsLinking(int linkingIndex)
//! Show the dual variable of the i-th constraint inside the linking constraints
{
  if (linkingIndex < 0 || linkingIndex >= numLinCons)
    cerr << "Index out of range" << endl;
  else
    cout << consNames[numCons-numLinCons+linkingIndex] << " " << y[numCons-numLinCons+linkingIndex] << endl;
}
