
#include <limits> // for numeric_limits
#include <stdlib.h> // for random
#include <algorithm> // for sort
#include <sstream>
#include <unistd.h> // for chdir
#include "SMLBlockIP.h"
#include "sml.h"
#include "ExceptionSMLBlockIP.h"

using namespace std;

void SMLBlockIP::fobjnonlin(int n, double x[], double &fx, double Gx[], double Hx[], void *params)
// Function to get objective function information in a point when the objective function is non-linear
/*!
  @param n Number of variables
  @param x Point
  @param fx Objective function value in x
  @param Gx Gradient in x
  @param Hx Hessian in x
  @param params User parameters to perform objective function calculations
*/
{
  objFunction *objF = (objFunction*) params;
  fx = 0.0;
  int index = 0;
  int numBlocks = objF->numBlocks;
  if (objF->slacks && objF->slacksOrder == NULL)
    ++numBlocks;
  for (int i = 0; i < numBlocks; ++i) {
    double LocalFx = 0.0;
    if (objF->blocks[i].nonlinear) {
      objF->blocks[i].model->getNonlinearObj(&x[index], LocalFx, &Gx[index], &Hx[index]);
    }
    else {
      for (int j = 0; j < objF->blocks[i].numVars; ++j) {
        LocalFx += objF->blocks[i].linearCoefs[j] * x[index+j];
        Gx[index+j] = objF->blocks[i].linearCoefs[j];
        Hx[index+j] = 0;
      }
      for (int j = 0; j < objF->blocks[i].numQuadVars; ++j) {
        LocalFx += objF->blocks[i].quadraticCoefs[j] * 0.5 * x[index+j] * x[index+j];
        Gx[index+j] += objF->blocks[i].quadraticCoefs[j] * x[index+j];
        Hx[index+j] += objF->blocks[i].quadraticCoefs[j];
      }
    }
    fx += LocalFx;
    index += objF->blocks[i].numVars;
  }
  if (objF->slacks && objF->slacksOrder != NULL) {
    double LocalFx = 0.0;
    if (objF->blocks[numBlocks].nonlinear) {
      int numLocalVars = objF->blocks[numBlocks].numVars;
      double *auxx = NULL, *auxGx = NULL, *auxHx = NULL;
      try {
        auxx = new double[numLocalVars];
        auxGx = new double[numLocalVars];
        auxHx = new double[numLocalVars];
        for (int i = 0; i < numLocalVars; ++i) auxx[i] = x[index+objF->slacksOrder[i]];
        objF->blocks[numBlocks].model->getNonlinearObj(auxx, LocalFx, auxGx, auxHx);
      } catch (exception& e) {
        delete[] auxx; delete[] auxGx; delete[] auxHx;
        throw;
      }
      for (int i = 0; i < numLocalVars; ++i) Gx[index+objF->slacksOrder[i]] = auxGx[i];
      for (int i = 0; i < numLocalVars; ++i) Hx[index+objF->slacksOrder[i]] = auxHx[i];
      delete[] auxx;
      delete[] auxGx;
      delete[] auxHx;
    }
    else {
      for (int j = 0; j < objF->blocks[numBlocks].numVars; ++j) {
        int k = objF->slacksOrder[j];
        LocalFx += objF->blocks[numBlocks].linearCoefs[j] * x[index+k];
        Gx[index+k] = objF->blocks[numBlocks].linearCoefs[j];
        Hx[index+k] = 0;
      }
      for (int j = 0; j < objF->blocks[numBlocks].numQuadVars; ++j) {
        int k = objF->slacksOrder[j];
        LocalFx += objF->blocks[numBlocks].quadraticCoefs[j] * 0.5 * x[index+k] * x[index+k];
        Gx[index+k] += objF->blocks[numBlocks].quadraticCoefs[j] * x[index+k];
        Hx[index+k] += objF->blocks[numBlocks].quadraticCoefs[j];
      }
    }
    fx += LocalFx;
  }
  if (!objF->slacks) {
    for (; index < n; ++index) {
      Gx[index] = 0;
      Hx[index] = 0;
    }
  }
}

SMLBlockIP::SMLBlockIP()
//! Constructor
{
  initialize();
}

/*
 * Each model block looks like the following:
 * A_1         C_1
 *     A_2     C_2
 *         A_3 C_3
 * B_1 B_2 B_3  D
 *
 * model->getJacobianOfIntersection(model) returns  D
 * model->getJacobianOfIntersection(A_i)   returns B_i
 * A_i->getJacobianOfIntersection(model)   returns C_i
 */
SMLBlockIP::SMLBlockIP(ExpandedModelInterface *root) {
  initialize();
  this->root = root;
  loadProblemInBlockIP();
}

SMLBlockIP::~SMLBlockIP()
//! Destructor
{
  freeMemory();
}

BlockIP* SMLBlockIP::getBlockIPInterface()
//! Get BlockIP interface
{
  return bip;
}

void SMLBlockIP::loadProblemInBlockIP()
//! Load a problem from SML structure to BlockIP
/*!
  
*/
{
  double constantFObj = 0;
  // get number of blocks, number of contraints and number of variables; also detect if a slacks block exists
  for(ExpandedModelInterface::child_iterator emit=root->cbegin(); emit!=root->cend(); ++emit) {
    ExpandedModelInterface *model = *emit;
    if (model->getNLocalVarsNonlinearCons() > 0)
      throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::NONLINEAR_CONSTRAINTS, __FILE__, __LINE__);
    numVars += model->getNLocalVars();
    numCons += model->getNLocalCons();
    if (model->getName() == "root_Slacks") {
      slacks = true;
      slacksBlock = model;
    }
    ++numBlocks;
  }
  --numBlocks; // global block that contains other blocks no considered
  
  nonLinear = false; // at least one block has an non-linear objective function (no quadratic, no linear)
  linear = true; // all blocks have an linear objective function (no quadratic, no non-linear)
  
  objF = new objFunction;
  objF->slacksOrder = NULL;
  objF->blocks = NULL;
  objF->blocks = new objBlock[numBlocks];
  for (int i = 0; i < numBlocks; ++i) {
    objF->blocks[i].linearCoefs = NULL;
    objF->blocks[i].quadraticCoefs = NULL;
  }
  
  {
    // detect type of objective function
    int objBlockIndex = 0;
    for(ExpandedModelInterface::child_iterator emit=root->cbegin(); objBlockIndex < numBlocks; ++emit, ++objBlockIndex) {
      ExpandedModelInterface *model = *emit;
      int numCoefQuad = 0;
      if (model->getNLocalNonlinearObj()) // quadratic or non-linear objective function
        numCoefQuad = model->getQuadraticObjNz();
      if (numCoefQuad  == -1) { // non-linear objective function
        nonLinear = true;
        objF->blocks[objBlockIndex].nonlinear = true;
      }
      else if (numCoefQuad < -1) {
        cerr << "There are a error in the objective function of " << model->getName() << "." << endl;
        throw;
      }
      else {
        objF->blocks[objBlockIndex].nonlinear = false;
        objF->blocks[objBlockIndex].numQuadVars = numCoefQuad; // only diagonal quadratic matrix allowed
        if (numCoefQuad > 0) linear = false;
      }
    }
  }
  
  numLinCons = root->getNLocalCons();
  
  if (slacks) {
    blocks = new MatrixBlockIP[numBlocks-1];
    linking_constraints = new MatrixBlockIP[numBlocks-1];
    objF->numBlocks = numBlocks-1;
    objF->slacks = true;
    blockNames = new string[numBlocks-1];
  }
  else {
    // add slacks variables
    numVars += numLinCons;
    blocks = new MatrixBlockIP[numBlocks];
    linking_constraints = new MatrixBlockIP[numBlocks];
    objF->numBlocks = numBlocks;
    objF->slacks = false;
    blockNames = new string[numBlocks];
  }
  
  { // Create MatrixBlockIP matrices
    ExpandedModelInterface::child_iterator emit=root->cbegin();
    for (int b = 0; b < numBlocks; ++b, ++emit) {
      ExpandedModelInterface *model = *emit;
      if (model != slacksBlock) {
        int *inicola;
        int *irowa;
        double *elements;
        //N block
        int m = model->getNLocalCons();
        int n = model->getNLocalVars();
        int nz = model->getNzJacobianOfIntersection(model);
        // vectors are ordered, it is not needed call order_matrix()
        blocks[b].create_general_matrix_column_wise(m, n, nz, inicola, irowa, elements);
        model->getJacobianOfIntersection(model, inicola, NULL, irowa, elements);
        
        // L block
        nz = root->getNzJacobianOfIntersection(model);
        // vectors are ordered, it is not needed call order_matrix()
        linking_constraints[b].create_general_matrix_column_wise(numLinCons, n, nz, inicola, irowa, elements);
        root->getJacobianOfIntersection(model, inicola, NULL, irowa, elements);
      }
    }
  }
  
  double *ub; // upper bounds
  double *lb; // lower bounds
  double *rhs; // right hand side
  double *lhs; // left hand side
  double *cost; // linear cost in the objective function
  double *qcost; // quadratic cost in the objective function
  
  { // Create BlockIP
    int n = numBlocks;
    if (slacks) --n;
    if (nonLinear){
      double *dummy;
      bip = new BlockIP(BlockIP::NONLINEAR, dummy, dummy, fobjnonlin, (void*)(objF), lb, ub, lhs, rhs, n, false, blocks, false, linking_constraints);
    }
    else {
      BlockIP::TYPE_PROBLEM type_objective = linear ? BlockIP::LINEAR : BlockIP::QUADRATIC;
      bip = new BlockIP(type_objective, cost, qcost, NULL, NULL, lb, ub, lhs, rhs, n, false, blocks, false, linking_constraints);
    }
  }

  int objBlockIndexGlobal = 0;
  int varsIndexGlobal = 0;
  int consIndex = 0;
  double bip_inf = bip->get_inf();
  ExpandedModelInterface::child_iterator emit=root->cbegin();
  for (int b = 0; b < numBlocks; ++b, ++emit) {
    ExpandedModelInterface *model = *emit;
    int numLocalVars = model->getNLocalVars();
    int numLocalCons = model->getNLocalCons();
    int objBlockIndex;
    int varsIndexLocal;
    if (model != slacksBlock) {
      objBlockIndex = objBlockIndexGlobal;
      varsIndexLocal = varsIndexGlobal;
    }
    else {
      objBlockIndex = numBlocks-1;
      varsIndexLocal = numVars - numLinCons;
    }
    objF->blocks[objBlockIndex].numVars = numLocalVars;
    objF->blocks[objBlockIndex].model = model;

    model->getColLowBounds(&lb[varsIndexLocal]);
    model->getColUpBounds(&ub[varsIndexLocal]);
    
    if (model == slacksBlock) {
      for (int i = 0; i < numLocalVars; ++i) {
        if (lb[varsIndexLocal+i] != 0) throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::NON_ZERO_SLACK_LB, __FILE__, __LINE__);
        if (ub[varsIndexLocal+i] < 0) throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::NEGATIVE_SLACK_UB, __FILE__, __LINE__);
      }
    }
    
    if (objF->blocks[objBlockIndex].nonlinear) { // nonlinear
      // check if hessian is semi-definit positive matrix in 2 random points,
      // this is non an absolute check but can detect almost all cases.
      if (!model->isHessianDiagonal())
        throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::HESS_NOT_DIAG);//, __FILE__, __LINE__);
        
      int numTest= 2;
      double *x = NULL;
      double *Hx = NULL;
      try {
        x = new double[numLocalVars];
        Hx = new double[numLocalVars];
        for (int j = 0; j < numTest; ++j) {
          for (int i = 0; i < numLocalVars; ++i) {
            double lwr = lb[varsIndexLocal+i];
            double upr = ub[varsIndexLocal+i];
            if (lwr <= -inf) lwr = -bip_inf;
            if (upr >= inf) upr = bip_inf;
            x[i] = lwr + rand() % int(upr-lwr+1);
          }
          double fx;
          model->getNonlinearObj(x, fx, NULL, Hx);
          for (int i = 0; i < numLocalVars; ++i) {
            if (Hx[i] < 0)
              throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::HESS_NOT_SEMIDEF);//, __FILE__, __LINE__);
          }
        }
      } catch (exception& e) {
        delete[] x; delete[] Hx;
        throw;
      }
      delete[] x;
      delete[] Hx;
      objF->blocks[objBlockIndex].quadraticCoefs = NULL;
      objF->blocks[objBlockIndex].linearCoefs = NULL;
    }
    else {
      double localConstant = 0;
      double *x = new double[numLocalVars];
      for (int i = 0; i < numLocalVars; ++i)
        x[i] = 0;
      if (nonLinear) {
        objF->blocks[objBlockIndex].linearCoefs = new double[numLocalVars];
        for (int i = 0; i < numLocalVars; ++i)
          objF->blocks[objBlockIndex].linearCoefs[i] = 0;  
        // model->getObjGradient(objF->blocks[objBlockIndex].linearCoefs);
        model->getNonlinearObj(x, localConstant, objF->blocks[objBlockIndex].linearCoefs, NULL);
      }
      else 
        // model->getObjGradient(&cost[varsIndexLocal]);
        model->getNonlinearObj(x, localConstant, &cost[varsIndexLocal], NULL);
      delete[] x;
      constantFObj += localConstant;
      
      int numCoefQuad = objF->blocks[objBlockIndex].numQuadVars;
      if (numCoefQuad == 0) {
        if (nonLinear)
          objF->blocks[objBlockIndex].quadraticCoefs = NULL;
      }
      else {
        int numLocalVarsNonlinearObj = model->getNLocalVarsNonlinearObj();
        // check quadratic diagonal matrix
        if (numCoefQuad > numLocalVarsNonlinearObj)
          throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::WRONG_QUADRATICS, __FILE__, __LINE__);
                    
        double *quadCoefsVals = NULL;
        if (nonLinear) {
          objF->blocks[objBlockIndex].quadraticCoefs = new double[numCoefQuad];
          quadCoefsVals = &objF->blocks[objBlockIndex].quadraticCoefs[0];
        }
        else
          quadCoefsVals = &qcost[varsIndexLocal];
        int *quadCoefsColsBeg = NULL;
        int *quadCoefsRows = NULL;  
        try {
          quadCoefsColsBeg = new int[numLocalVarsNonlinearObj+1];
          quadCoefsRows = new int[numCoefQuad];
          model->getQuadraticObj(quadCoefsColsBeg, quadCoefsRows, quadCoefsVals, true);
          // check quadratic diagonal positive matrix
          for (int i = 0; i < numLocalVarsNonlinearObj; ++i)
            if (quadCoefsColsBeg[i+1] - quadCoefsColsBeg[i] != 1)
              throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::WRONG_QUADRATICS, __FILE__, __LINE__);
          for (int i = 0; i < numCoefQuad-1; ++i)
            if (quadCoefsRows[i+1] - quadCoefsRows[i] != 1)
              throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::WRONG_QUADRATICS, __FILE__, __LINE__);
          for (int i = 0; i < numCoefQuad; ++i)
            if (quadCoefsVals[i] < 0)
              throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::WRONG_QUADRATICS, __FILE__, __LINE__);
        } catch (exception& e) {
          delete[] quadCoefsColsBeg; delete[] quadCoefsRows;
          throw;
        }
        delete[] quadCoefsColsBeg;
        delete[] quadCoefsRows;
        if (!nonLinear)
          for (int i = numCoefQuad; i < numLocalVars; ++i)
            quadCoefsVals[i] = 0;
      }
    }
    
    if (model != slacksBlock) {
      model->getRowBounds(&lhs[consIndex], &rhs[consIndex]);
      ++objBlockIndexGlobal;
      varsIndexGlobal += numLocalVars;
      consIndex += numLocalCons;
    }
    else {
      int nz = root->getNzJacobianOfIntersection(model);
      int *inicola = NULL;
      int *irowa = NULL;
      double *elements = NULL;
      try {      
        inicola = new int[numLocalVars+1];
        irowa = new int[nz];
        elements = new double[nz];
        root->getJacobianOfIntersection(model, inicola, NULL, irowa, elements);
      } catch (exception& e) {
        delete[] inicola; delete[] irowa; delete[] elements;
        throw;
      }
      if (nz != numLocalVars)
        throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::LIN_SLACKS, __FILE__, __LINE__);
      for (int i = 0; i < nz; ++i)
        if (elements[i] != 1) throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::WRONG_SLACK_COEF, __FILE__, __LINE__);
      bool ordered = true;
      for (int i = 0; i < numLocalVars; ++i)
        if (irowa[i] != i) {
          ordered = false;
          break;
        }
      if (ordered) {
        objF->slacksOrder = NULL;
        delete[] irowa;
      }
      else {
        double *aux1 = NULL;
        double *aux2 = NULL;
        try {
          aux1 = new double[numLocalVars];
          aux2 = new double[numLocalVars];
        } catch (exception& e) {
          delete[] irowa; delete[] inicola; delete[] elements;
          delete[] aux1; delete[] aux2;
          throw;
        }
        for (int i = 0; i < numLocalVars; ++i) {
          int k = irowa[i];
          aux1[k] = ub[varsIndexLocal+i];
          aux2[k] = lb[varsIndexLocal+i];
        }
        for (int i = 0; i < numLocalVars; ++i) {
          ub[varsIndexLocal+i] = aux1[i];
          lb[varsIndexLocal+i] = aux2[i];
        }
        if (!nonLinear) {
          for (int i = 0; i < numLocalVars; ++i) {
            int k = irowa[i];
            aux1[k] = cost[varsIndexLocal+i];
          }
          for (int i = 0; i < numLocalVars; ++i) {
            cost[varsIndexLocal+i] = aux1[i];
          }
          if (!linear) {
            for (int i = 0; i < numLocalVars; ++i) {
              int k = irowa[i];
              aux2[k] = qcost[varsIndexLocal+i];
            }
            for (int i = 0; i < numLocalVars; ++i) {
              qcost[varsIndexLocal+i] = aux2[i];
            }
          }
          delete[] irowa;
        }
        else
          objF->slacksOrder = irowa;
        delete[] aux1;
        delete[] aux2;
      }
      delete[] inicola;
      delete[] elements;
    }
  }

  root->getRowBounds(&lhs[consIndex], &rhs[consIndex]);
  if (slacks) {
    for (int k = 0; k < root->getNLocalCons(); ++k) {
      if (lhs[consIndex+k] != rhs[consIndex+k])
        throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::NON_EQUAL_COMMODITIES, __FILE__, __LINE__);
    }
  }
  else {
    int j = 0;
    for (int i = numVars - numLinCons; i < numVars; ++i, ++j) {
      lb[i] = 0;
      ub[i] = 0;
      if (!nonLinear) {
        cost[i] = 0;
        if (!linear)
          qcost[i] = 0;
      }
    }
  }
  
  {
    if (!nonLinear) {
      for (int i = 0; i < numBlocks; ++i) {
        delete[] objF->blocks[i].linearCoefs;
        delete[] objF->blocks[i].quadraticCoefs;
      }
      delete[] objF->blocks;
      delete[] objF->slacksOrder;
      delete objF;
      objF = NULL;
    }
    bip->set_constant_fobj(bip->get_constant_fobj()+constantFObj);
    {
      int *unorderedVars = NULL;
      try {
        unorderedVars = new int[numVars];
        for (int i = 0; i < numVars; ++i)
          unorderedVars[i] = i;
        
        varNames = new string[numVars];
        varsOrder = new int[numVars];
        consNames = new string[numCons];
        int blockCount = 0;
        int varCount = 0;
        int consCount = 0;
        ExpandedModelInterface::child_iterator emit=root->cbegin();
        for (int b = 0; b < numBlocks; ++b, ++emit) {
          ExpandedModelInterface *model = *emit;
          int LocalVarCount;
          int LocalConsCount;
          if (model == slacksBlock) {
            LocalVarCount = numVars - numLinCons;
            LocalConsCount = numCons - numLinCons;
          }
          else {
            LocalVarCount = varCount;
            LocalConsCount = consCount;
            blockNames[blockCount] = model->getName();
            blockNames[blockCount].erase(0,5);
            ++blockCount;
          }
          int numLocalVars = model->getNLocalVars();
          int numLocalCons = model->getNLocalCons();
          list<string> varNamesList = model->getLocalVarNames();
          list<string>::iterator itVarNames = varNamesList.begin();
          list<string> consNamesList = model->getLocalConNames();
          list<string>::iterator itConsNames = consNamesList.begin();
          string* *aux_varNames = NULL;
          // v will contains the correct order
          vector<int> *v = NULL;
          try {
            aux_varNames = new string*[numLocalVars];
            v = new vector<int>;
            for (int i = 0; i < numLocalVars; ++i, ++itVarNames)
              aux_varNames[i] = (string*)(&(*itVarNames));

            Order_vector order;
            v->resize(numLocalVars);
            for (int i= 0; i < numLocalVars; ++i) (*v)[i]= i;
            order.v= aux_varNames;
            
            // Sort in this way does not work properly
            // sort(v->begin(), v->end(), order);
            
            // puts the vectors in the correct order
            for (int i= 0; i < numLocalVars; ++i) {
              varNames[LocalVarCount+i] = *aux_varNames[(*v)[i]];
              varsOrder[LocalVarCount+i] = unorderedVars[LocalVarCount+(*v)[i]];
            }
          } catch(exception& e) {
            delete[] aux_varNames; delete v;
            throw;
          }
          delete[] aux_varNames;
          delete v;
          varCount += numLocalVars;
          consCount += numLocalCons;
          
          for (int i= 0; i < numLocalCons; ++i, ++itConsNames) {
            consNames[LocalConsCount+i] = *itConsNames;
          }
        }
        if (!slacks) {
          int numSlack = 1;
          for (int i = numVars - numLinCons; i < numVars; ++i, ++numSlack) {
            stringstream ss;
            ss << "Slacks[" << numSlack << "]";
            ss >> varNames[i];
            varsOrder[i] = i;
          }
        }
        if (root->getNLocalCons() > 0) {
            list<string> consNamesList = root->getLocalConNames();
            list<string>::iterator itConsNames = consNamesList.begin();
            for (int i = numCons - numLinCons; i < numCons; ++i, ++itConsNames) {
              consNames[i] = *itConsNames;
          }
        }
      } catch (exception& e) {
        delete[] unorderedVars;
        throw;
      }
      delete[] unorderedVars;
    }
    
    string *unorderedVarNames = new string[numVars];
    for (int i = 0; i < numVars; ++i)
      unorderedVarNames[varsOrder[i]] = varNames[i];
    bip->set_names(blockNames, unorderedVarNames, consNames, true);
    delete[] unorderedVarNames;
  }
  
  int n_vars= bip->n_vars;
  int m_cons= bip->m_cons;
  for (int i = 0; i < n_vars; ++i) {
    if (ub[i] >= bip_inf) {
      ub[i] = bip_inf;
    }
    if (lb[i] <= -inf)
      lb[i] = -bip_inf;
  }
  
  for (int i = 0; i < m_cons; ++i) {
    if (rhs[i] >= bip_inf) {
      rhs[i] = bip_inf;
    }
    if (lhs[i] <= -inf)
      lhs[i] = -bip_inf;
  }
}

void SMLBlockIP::readAmpl(const char *modelFilename, const char *dataFilename)
//! Load a problem from ampl files
/*!
  
*/
{
  freeMemory();
  initialize();
  root = sml_generate(string(modelFilename), string(dataFilename), false);
  if (!root) throw ExceptionSMLBlockIP(ExceptionSMLBlockIP::SYNTAX_ERROR);
  loadProblemInBlockIP();
#ifdef WIN32
  int errcode = _chdir("..");
#else
  int errcode = chdir("..");
#endif
  if (errcode){
    cerr << "ERROR: Failed to change working directory to 'tmp/'\n";
    throw;
  }
}

void SMLBlockIP::amplToMps(const char *modelFilename, const char *dataFilename, const char *mpsFilename, bool convertToStd)
//! Converts a problem in ampl format to mps format
/*!
  @param convertToStd If true the problem will be written in BlockIP standard form
  @note The problem contained by BlockIP will be not modified
*/
{
  SMLBlockIP *smlbip = new SMLBlockIP();
  try {
    smlbip->readAmpl(modelFilename, dataFilename);
    if (convertToStd)
      smlbip->getBlockIPInterface()->convert_to_std_and_write_mps(mpsFilename);
    else
      smlbip->getBlockIPInterface()->write_mps(mpsFilename);
  } catch (exception &e) {
    delete smlbip;
    throw;
  }
  delete smlbip;
}

void SMLBlockIP::amplToBlockIPFormat(const char *modelFilename, const char *dataFilename, const char *BlockIPFormatFilename, bool convertToStd)
//! Converts a problem in ampl format to BlockIP format
/*!
  @param convertToStd If true the problem will be written in BlockIP standard form
*/
{
  SMLBlockIP *smlbip = new SMLBlockIP();
  try {
    smlbip->readAmpl(modelFilename, dataFilename);
    if (convertToStd)
      smlbip->getBlockIPInterface()->convert_to_standard();
    smlbip->getBlockIPInterface()->write_BlockIP_format(BlockIPFormatFilename);
  } catch (exception &e) {
    delete smlbip;
    throw;
  }
  delete smlbip;
}

void SMLBlockIP::printProblem()
//! Print the whole problem throw standard output
{
  if (!bip) return;
  int n_vars= bip->n_vars;
  int m_cons= bip->m_cons;
  int n = numBlocks;
  if (slacks) --n;    
  cout << endl << "Diagonal blocks: " << endl;
  // for (int i = 0; i < n; ++i) blocks[i].column_wise_to_row_wise_format(); // TEST ROW-WISE CONVERSION
  for (int i = 0; i < n; ++i) blocks[i].print_matrix(false,false);
  cout << endl << "Linking constraints: " << endl;
  // for (int i = 0; i < n; ++i) linking_constraints[i].column_wise_to_row_wise_format(); // TEST ROW-WISE CONVERSION
  for (int i = 0; i < n; ++i) linking_constraints[i].print_matrix(false,false);
  
  cout << endl << "ub: ";
  for (int i = 0; i < n_vars; ++i) cout << bip->ub[i] << " ";
  cout << endl << "rhs: ";
  for (int i = 0; i < m_cons; ++i) cout << bip->rhs[i] << " ";
  cout << endl << endl;

  if (nonLinear) {
    double *x = NULL;
    double *Gx = NULL;
    double *Hx = NULL;
    double fx;
    try {
      x = new double[n_vars];
      for (int i = 0; i < n_vars; ++i) x[i] = 1;
      Gx = new double[n_vars];
      Hx = new double[n_vars];
      double *ext_x= NULL;
      bip->stdForm->original_to_transformed_primal_variables(ext_x, x);
      bip->stdForm->fobj_stdform(n_vars, ext_x, fx, Gx, Hx, bip->params);
      delete[] ext_x;
    } catch (exception& e) {
      delete[] x; delete[] Gx; delete[] Hx;
      throw;
    }
    delete[] x;
    cout << "fx: " << fx << endl;
    cout << "Gx: ";
    for (int i = 0; i < n_vars; ++i) cout << Gx[i] << " ";
    cout << endl << "Hx: ";
    for (int i = 0; i < n_vars; ++i) cout << Hx[i] << " ";
    cout << endl;
    delete[] Gx;
    delete[] Hx;
  }
  else {
    cout << "Constant fobj: " << bip->constantFObj << endl;
    cout << "cost: ";
    for (int i = 0; i < n_vars; ++i) cout << bip->cost[i] << " ";
    if (!linear) {
      cout << endl << "qcost: ";
      for (int i = 0; i < n_vars; ++i) cout << bip->qcost[i] << " ";
    }
    cout << endl;
  }
}

void SMLBlockIP::freeMemory() {
  if (objF) {
    if (objF->blocks) {
      for (int i = 0; i < numBlocks; ++i) {
        delete[] objF->blocks[i].linearCoefs;
        delete[] objF->blocks[i].quadraticCoefs;
      }
      delete[] objF->blocks;
    }
    delete[] objF->slacksOrder;
    delete objF;
    objF = NULL;
  }
  delete[] blocks;
  delete[] linking_constraints;
  blocks = NULL;
  linking_constraints = NULL;
  delete bip;
  bip = NULL;
  delete[] blockNames;
  blockNames = NULL;
  delete[] varNames;
  varNames = NULL;
  delete[] varsOrder;
  varsOrder = NULL;
  delete[] consNames;
  consNames = NULL;
}

void SMLBlockIP::initialize() {
  inf = numeric_limits<double>::infinity();
  numBlocks = 0;
  numVars = 0;
  numCons = 0;
  numLinCons = 0;
  slacks = false;
  nonLinear = false;
  linear = false;
  root = NULL;
  slacksBlock = NULL;
  objF = NULL;
  bip = NULL;
  blocks = NULL;
  linking_constraints = NULL;
  blockNames = NULL;
  varNames = NULL;
  varsOrder = NULL;
  consNames = NULL;
}
