/*!
  @file SMLBlockIP.h 
  @brief Definition of SMLBlockIP
*/

#ifndef SMLBlockIP_H
#define SMLBlockIP_H

#include <string>
#include "ExpandedModelInterface.h"
#include "BlockIP.h"
#include "MatrixBlockIP.h"

using namespace std;

//! Class for input problems in SML format to BlockIP
class SMLBlockIP {
  private:
    struct objBlock {
      bool nonlinear;
      int numVars;
      int numQuadVars;
      double *linearCoefs;
      double *quadraticCoefs;
      ExpandedModelInterface *model;
    };

    struct objFunction {
      int numBlocks;
      bool slacks;
      int *slacksOrder;
      objBlock *blocks;
    };
    
    struct Order_vector {
      string* *v;
      bool operator() (int i,int j) {
        return (*v[i] < *v[j]);
      }
    };
  
    int numBlocks; //!< blocks number including slacks
    int numVars; //!< number of variables including slacks
    int numCons; //!< number of constraints including linking constraints
    int numLinCons; //!< number of linking constraints
    bool slacks; //!< exist slacks block in ampl model
    bool nonLinear; //!< objective function is nonlinear
    bool linear; //!< objective function is linear
    ExpandedModelInterface *root; //!< root block
    ExpandedModelInterface* slacksBlock; //!< slacks block
    objFunction *objF; //!< to compute nonlinear objective functions
    BlockIP *bip; //!< BlockIP environment
    string *blockNames; //!< block names
    string *varNames; //!< variable names
    int *varsOrder; //!< correct order for the variables, used for output, ampl orders the variables
    string *consNames; //!< constraint names
    
    /*! These structures belong to BlockIP after its creation, then they are read-only */
    MatrixBlockIP *blocks; //!< main blocks
    MatrixBlockIP *linking_constraints; //!< linking constraints blocks
    
    double inf; //!< infinity
    
    static void fobjnonlin(int n, double x[], double &fx, double Gx[], double Hx[], void *params); //!< Function to get objective function information in a point when the objective function is non-linear
    
  public:
    // Constructor
    SMLBlockIP();
    
    // Constructor
    SMLBlockIP(ExpandedModelInterface *root);
    
    // Destructor
    ~SMLBlockIP();
    
    // Load a problem from ampl files
    void readAmpl(const char *modelFilename, const char *dataFilename);
    
    // Converts a problem in ampl format to mps format
    void amplToMps(const char *modelFilename, const char *dataFilename, const char *mpsFilename, bool convertToStd=false);
    
    // Converts a problem in ampl format to BlockIP format
    void amplToBlockIPFormat(const char *modelFilename, const char *dataFilename, const char *BlockIPFormatFilename, bool convertToStd=false);
    
    // Get BlockIP interface
    BlockIP* getBlockIPInterface();
    
    // Print the whole problem through standard output
    void printProblem();
        
    // Return true if the objective function is non-linear
    bool isNonLinear();
    
    // Return true if the objective function is quadratic
    bool isQuadratic();
    
    // Return the correct order of the variables, ampl change the order
    const int* getVarsOrder();
    
  private:
    void initialize();
    
    // Load a problem from SML structure to BlockIP
    void loadProblemInBlockIP();
    
    void freeMemory();
};

inline bool SMLBlockIP::isNonLinear()
//! Return true if the objective function is non-linear
/*!
  @return true if the objective function is non-linear
*/
{
  return nonLinear;
}

inline bool SMLBlockIP::isQuadratic()
//! Return true if the objective function is quadratic
/*!
  @return true if the objective function is quadratic
*/
{
  return !linear && !nonLinear;
}

inline const int* SMLBlockIP::getVarsOrder()
//! Return the correct order of the variables, ampl change the order
/*!
  @return The correct order of the variables
*/
{
  return varsOrder;
}

#endif //SMLBlockIP_H