/*!
  @file mainBlockIPPlatform.cpp
  @brief Command line program to solve large block angular problems

  The program has two different usages, solve a problem and convert a
  problem from one input format to other.
  
  The input problem can be in sml extended ampl files, in mps extended to blocks
  or in BlockIP format. 
  The blocks structure must be angular and the quadratic terms in the objective
  function cannot be given by two different variables. For more information
  about the input files read the SMLBlockIP and BlockIP documentation.
  
  The only solver that take advantage of the block structure is BlockIP.
  
  usage: BlockIPPlatform mode [options] [additional options]
  where mode is:
    If you want to solve a problem must choose an input format:
    -ampl {model filename} {data filename}
      Input problem in sml extended ampl format.
    -mps {mps filename}
      Input problem in mps extended to blocks format.
    -BlockIPFormat {BlockIPFormat filename}
      Input problem in BlockIP format.
      
    If you want to convert a problem from one input format to other:
    -amplToMps {modelFilename} {dataFilename} {mpsFilename}
      Converts a problem in ampl format to mps format
    -amplToBlockIPFormat {modelFilename} {dataFilename} {BlockIPFormatFilename}
      Converts a problem in ampl format to BlockIP format
    -mpsToBlockIPFormat {mpsFilename} {BlockIPFormatFilename}
      Converts a problem in mps format to BlockIP format
    -BlockIPFormatToMps {BlockIPFormatFilename} {mpsFilename}
      Converts a problem in BlockIP format to mps format
  
  options for convert mode
    -standarize
      Standarize the problem before write the output file converting all
      restrictions in equalities and all variables with 0 lower bound.
  
  additional options for convert mode
    -outDir {path}
      Set the directory where the output files will we written, must exist,
      by default it is the actual directory.
      
  options for solve mode
    Optionally you can set some options:
    -solver {solver}
      Set the solver to use to solve the problem. Depends of your installation
      but are allowed blockip, blockip_matlab, cplex, xpress, glpk and clp, by default blockip
      is used.
    -outDir {path}
      Set the directory where the output files will we written, must exist,
      by default it is the actual directory.
    -logFile {filename}
      Set the name of the log file, by default the same that input file with
      extension .log.
    -solFile {filename}
      Set the name of the solution file, by default the same that input file
      with extension .sol.
    -time {number of seconds}
      Set the maximum number of seconds that the solver can use to solve the
      problem.
    -threads {number of threads to use}
      Set the maximum number of threads that the solver can use to solve the
      problem, by default unlimited.
    -pathMatlab {path}
      If you want to use the Matlab implementation of BlockIP you need prblock_ip.m
      file. The default directory where is expected to find it is "../prblock_ip".
    
  additional for solve mode
    If the solver to use is BlockIP, you can set addicional options, for the 
    default values see the constant terms in BlockIP.h. blockip_matlab only 
    accepts -in, -iter and -type_comp_dy options 
    -interactive
      After solving the problem starts a interactive mode to retrieve information
    -inf {value}
      Infinity value.
    -m_pw_prec {value}
      Set number of terms used as preconditioner of the power series expansion of (D-C'*B^{-1}*B)^{-1}
    -sigma {value}
      Reduction of the centrality parameter at each IP iteration.
    -rho {value}
      Reduction of the step-length for the primal and dual variables at each 
      -IP iteration.
    -optim_gap {value}
      Optimality gap tolerance.
    -optim_pfeas {value}
      Primal feasibility tolerance.
    -optim_dfeas {value}
      Dual feasibility tolerance.
    -output_freq {value}
      Output information lines will be printed each selected IP -iterations.
    -output {value}
      Set type of output. Can take the values: 0= NONE, 1= SCREEN, 2= FILE, 3=BOTH
    -maxiter {value}
      Maximum number of IP iterations.
    -min_pcgtol {value}
      Minimum tolerance for the conjugate gradient.
    -red_pcgtol {value}
      Reduction of pcg_tol at each IP iteration.
    -init_pcgtol {value}
      Initial tolerance for the conjugate gradient (by default PCG_TOL_LIN if
      linear problem, PCG_TOL_QUAD if quadratic or -convex- nonlinear).
    -type_start_point {value}
      Set how starting point is computed. Can take the values: 0= SIMPLE, 1= QUAD_EQCONS_PROB
    -type_direction {value}
      Set which direction is computed. Can take the values: 0= NEWTON, 1= SECOND_ORDER, 2= AUTOMATIC
    -deactivateLnk {value}
      Set flag on deactivation of linking
    -iter_pcg {value}
      Maximum number of pcg iterations (if 0, then the value will be computed by 
      the code).
    -type_reg {value}
      Set type of regularization performed. Can take the values: 0=NO_REG, 1= QUAD_REG, 2= PROX_REG
    -factor_reg {value}
      Set initial value for regularization term
    -show_specrad {y/n}
      Whether estimation of spectral radius by Ritz values is shown at each IP iteration
    -type_comp_dy {value}
      Scalar that can take values:
      0: (default value) the dy direction is computed with Cholesky for blocks,
         and PCG for linking constraints 
      1: the dy direction is computed with Cholesky for the full A*Theta*A' system.
      2: Hybrid Preconditioner: Phase 1) the dy direction is computed with Cholesky
         for blocks, and PCG for linking constraints and Phase 2) dy direction is 
         computed with pcg using the split preconditioner for the A*Theta*A' system. 
      3: Theta0 preconditioner: like type_comp_dy=2, but using Theta0 (Theta of
         slacks of linking constraints) instead of power series preconditioner. 
         This may be useful when the space of linking constraints is "close" 
         (principal angles not large) to the space of block constraints.
  
  Interactive mode
    help    Show short help
    help -v    Show extended help
    pobj    Show primal objective function
    dobj    Show dual objective function
    it    Show the number of IP iterations performed
    cgit    Show the overall number of PCG iterations performed
    exit    Ends the interactive mode

    Indices start in 0.
    Display the primal variables value of blocks and slacks
    x
      Show all the primal variables of blocks and slacks
    x {i}    Example: x 5
      Show the value of the i-th primal variable
    x {name}    Example: x Block1:Var5
      Show the value of the named primal variable
    x {k} {i}    Example: x 1 5
      Show the value of the i-th primal variable inside the block k
    x {blockName} {i}    Example: x Block1 5
      Show the value of the i-th primal variable inside the named block
    x block {k}    Example: x block 1
      Show the value of all primal variables inside the block k
    x block {blockName}    Example: x block Block1
      Show the value of all primal variables inside the named block
    x slack    Example: x slack
      Show the value of all primal variables inside the slacks
    x slack {i}    Example: x slack 3
      Show the value of the i-th primal variable inside the slacks

    Display the reduced cost of primal variables of blocks and slacks
    rc
      Show all the reduced cost of primal variables of blocks and slacks
    rc {i}    Example: rc 5
      Show the reduced cost of the i-th primal variable
    rc {name}    Example: rc Block1:Var5
      Show the reduced cost of the named primal variable
    rc {k} {i}    Example: rc 1 5
      Show the reduced cost of the i-th primal variable inside the block k
    rc {blockName} {i}    Example: rc Block1 5
      Show the reduced cost of the i-th primal variable inside the named block
    rc block {k}    Example: rc block 1
      Show the reduced cost of all primal variables inside the block k
    rc block {blockName}    Example: rc block Block1
      Show the reduced cost of all primal variables inside the named block
    rc slack    Example: rc slack
      Show the reduced cost of all primal variables inside the slacks
    rc slack {i}    Example: rc slack 3
      Show the reduced cost of the i-th primal variable inside the slacks
    
    Display the dual variables value of blocks and linking constraints
    cons
      Show all the dual variables of blocks and linking constraints
    cons {i}    Example: cons 5
      Show the dual variable of the i-th constraint
    cons {name}    Example: cons Block1:Cons5
      Show the dual variable of the named constraint
    cons {k} {i}    Example: cons 1 5
      Show the dual variable of the i-th constraint inside the block k
    cons {blockName} {i}    Example: cons Block1 5
      Show the dual variable of the i-th constraint inside the named block
    cons block {k}    Example: cons block 1
      Show all the dual variables inside the block k
    cons block {blockName}    Example: cons block Block1
      Show all the dual varialbes inside the named block
    cons linking    Example: cons linking
      Show all the dual varialbes inside the linking constraints
    cons linking {i}    Example: cons linking 3
      Show the dual variable of the i-th constraint inside the linking constraints
*/

#include <iostream>
#include "BlockIPInterface.h"
#include "BlockIP.h"
#include <cstring>
#include <limits> // for numeric_limits
#include <sstream>

#ifdef WIN32
#define DIRSEP '\\'
#else
#define DIRSEP '/'
#endif

using namespace std;

void writeUsageMessage();
void writeHelp();
void writeExtHelp();

int main(int argc, char *argv[])
{
  enum MODE{NONE, SOLVE, CONVERT} mode = NONE;
  enum SUBMODE{BLOCKIPFORMAT, AMPL, MPS, AMPLTOMPS, AMPLTOBLOCKIPFORMAT, MPSTOBLOCKIPFORMAT, BLOCKIPFORMATTOMPS} submode;
  if (argc == 1) {
    writeUsageMessage();
    return 1;
  }
  // cout.precision(14);
  char *modFile = NULL;
  char *datFile = NULL;
  char *mpsFile = NULL;
  char *BlockIPFormatFile = NULL;
  char *outFile = NULL;
  BlockIPInterface::SOLVER solver = BlockIPInterface::BLOCKIP;
  
  char *outDir = NULL;
  char *logFile = NULL;
  char *solFile = NULL;
  double time = -1;
  int threads = -1;
  bool standarize = false;
  char *pathMatlab = NULL;
  int argIndex;
  bool interactive = false;
  if (!strcmp(argv[1], "-ampl")) {
    if (argc < 4) {
      writeUsageMessage();
      return 1;
    }
    modFile = argv[2];
    datFile = argv[3];
    argIndex = 4;
    mode = SOLVE;
    submode = AMPL;
  }
  else if (!strcmp(argv[1], "-mps")) {
    if (argc < 3) {
      writeUsageMessage();
      return 1;
    }
    mpsFile = argv[2];
    argIndex = 3;
    mode = SOLVE;
    submode = MPS;
  }
  else if (!strcmp(argv[1], "-BlockIPFormat")) {
    if (argc < 3) {
      writeUsageMessage();
      return 1;
    }
    BlockIPFormatFile = argv[2];
    argIndex = 3;
    mode = SOLVE;
    submode = BLOCKIPFORMAT;
  }
  else if (!strcmp(argv[1], "-amplToMps")) {
    if (argc < 5) {
      writeUsageMessage();
      return 1;
    }
    modFile = argv[2];
    datFile = argv[3];
    outFile = argv[4];
    argIndex = 5;
    mode = CONVERT;
    submode = AMPLTOMPS;
  }
  else if (!strcmp(argv[1], "-amplToBlockIPFormat")) {
    if (argc < 5) {
      writeUsageMessage();
      return 1;
    }
    modFile = argv[2];
    datFile = argv[3];
    outFile = argv[4];
    argIndex = 5;
    mode = CONVERT;
    submode = AMPLTOBLOCKIPFORMAT;
  }
  else if (!strcmp(argv[1], "-mpsToBlockIPFormat")) {
    if (argc < 4) {
      writeUsageMessage();
      return 1;
    }
    mpsFile = argv[2];
    outFile = argv[3];
    argIndex = 4;
    mode = CONVERT;
    submode = MPSTOBLOCKIPFORMAT;
  }
  else if (!strcmp(argv[1], "-BlockIPFormatToMps")) {
    if (argc < 4) {
      writeUsageMessage();
      return 1;
    }
    BlockIPFormatFile = argv[2];
    outFile = argv[3];
    argIndex = 4;
    mode = CONVERT;
    submode = BLOCKIPFORMATTOMPS;
  }
  else {
    writeUsageMessage();
    return 1;
  }
  
  if (mode == SOLVE) {
    for (; (argIndex+1) < argc; argIndex += 2) {
      if (!strcmp(argv[argIndex], "-solver")) {
        if (!strcmp(argv[argIndex+1], "blockip"))
          solver = BlockIPInterface::BLOCKIP;
        else if (!strcmp(argv[argIndex+1], "blockip_matlab"))
          solver = BlockIPInterface::BLOCKIP_MATLAB;
        else if (!strcmp(argv[argIndex+1], "cplex"))
          solver = BlockIPInterface::CPLEX;
        else if (!strcmp(argv[argIndex+1], "xpress"))
          solver = BlockIPInterface::XPRESS;
        else if (!strcmp(argv[argIndex+1], "glpk"))
          solver = BlockIPInterface::GLPK;
        else if (!strcmp(argv[argIndex+1], "clp"))
          solver = BlockIPInterface::CLP;
        else {
          writeUsageMessage();
          return 1;
        }
      }
      else if (!strcmp(argv[argIndex], "-outDir"))
        outDir = argv[argIndex+1];
      else if (!strcmp(argv[argIndex], "-logFile"))
        logFile = argv[argIndex+1];
      else if (!strcmp(argv[argIndex], "-solFile"))
        solFile = argv[argIndex+1];
      else if (!strcmp(argv[argIndex], "-time")) {
        sscanf(argv[argIndex+1], "%lf", &time);
        if (time <= 0)
          time = -1;
      }
      else if (!strcmp(argv[argIndex], "-threads")) {
        sscanf(argv[argIndex+1], "%i", &threads);
        if (threads < 0)
          threads = -1;
      }
      else if (!strcmp(argv[argIndex], "-pathMatlab"))
        pathMatlab = argv[argIndex+1];
      else
        break;
    }
  }
  else if (mode == CONVERT) {
    if (argc > argIndex) {
      if (!strcmp(argv[argIndex], "-standarize")) {
        standarize = true;
        ++argIndex;
      }
      else {
        writeUsageMessage();
        return 1;
      }
    }
    if (argc > argIndex) {
      if (!strcmp(argv[argIndex], "-outDir"))
        outDir = argv[argIndex+1];
      else {
        writeUsageMessage();
        return 1;
      }
    }
  }
  
  BlockIPInterface *bipi = NULL;
  try {
    bipi = new BlockIPInterface();
    if (mode == CONVERT) {
      if (outDir) {
        char *filename = outFile;
        size_t len1 = strlen(outDir);
        size_t len2 = strlen(filename);
        try {
          outFile = new char[len1+len2+2];
          sprintf(outFile, "%s%c%s", outDir, DIRSEP, filename);
        } catch (exception &e) {
          delete[] outFile;
          throw;
        }
      }
      if (submode == AMPLTOMPS)
        bipi->amplToMps(modFile, datFile, outFile, standarize);
      else if (submode == AMPLTOBLOCKIPFORMAT)
        bipi->amplToBlockIPFormat(modFile, datFile, outFile, standarize);
      else if (submode == MPSTOBLOCKIPFORMAT)
        bipi->mpsToBlockIPFormat(mpsFile, outFile, standarize);
      else if (submode == BLOCKIPFORMATTOMPS)
        bipi->BlockIPFormatToMps(BlockIPFormatFile, outFile, standarize);
      
      delete bipi;
      if (outDir)
        delete[] outFile;
        
      if (submode == AMPLTOMPS || submode == AMPLTOBLOCKIPFORMAT) {
#ifdef WIN32
        system("rd /S /Q tmp");
#else
        system("/bin/rm -r tmp");
#endif
      }
      
      cout << "Conversion performed successfully" << endl;
      return 0;
    }
    // mode == SOLVE
    if (solver == BlockIPInterface::BLOCKIP_MATLAB) {
      if (pathMatlab)
        bipi->pathMatlab = pathMatlab;
      else
        bipi->pathMatlab = "../prblock_ip";
    }
    
    bipi->setSolver(solver);
    
    if (time != -1)
      bipi->setMaxTime(time);
    char *filename = NULL;
    try {
      if (!logFile || ! solFile) { // a name must be created for the output files
        if (submode == BLOCKIPFORMAT)
          filename = BlockIPFormatFile;
        else if (submode == MPS)
          filename = mpsFile;
        else if (submode == AMPL)
          filename = modFile;
          
        char *initName, *endName;
        initName = strrchr(filename, DIRSEP);
        endName = strrchr(filename, '.');
        if (initName)
          initName = &(initName)[1];
        else
          initName = filename;
        if (!endName) endName = filename + sizeof(filename);
        char *aux = new char[endName-initName+1];
        memcpy(aux, initName, endName-initName);
        aux[endName-initName] = '\0';
        filename = aux;
      }
      if (logFile) {
        if (outDir) {
          char *aux = NULL;
          try {
            aux = new char[strlen(outDir)+strlen(logFile)+2];
            sprintf(aux, "%s%c%s", outDir, DIRSEP, logFile);
            logFile = aux;
            bipi->setLogFilename(logFile);
          } catch (exception &e) {
            delete[] aux;
            throw;
          }
          delete[] aux;
        }
        else
          bipi->setLogFilename(logFile);
      }
      else {
        char *logFile = NULL;
        try {
          if (outDir) {
            logFile = new char[strlen(outDir)+strlen(filename)+6];
            sprintf(logFile, "%s%c%s.log", outDir, DIRSEP, filename);
          }
          else {
            logFile = new char[strlen(filename)+5];
            sprintf(logFile, "%s.log", filename);
          }
          bipi->setLogFilename(logFile);
        } 
        catch (exception &e) {
          delete[] logFile;
          throw;
        }
        delete[] logFile;
      }
      if (solFile) {
        if (outDir) {
          char *aux = NULL;
          try {
            aux = new char[strlen(outDir)+strlen(solFile)+2];
            sprintf(aux, "%s%c%s", outDir, DIRSEP, solFile);
            solFile = aux;
            bipi->setSolFilename(solFile);
          } catch (exception &e) {
            delete[] aux;
            throw;
          }
          delete[] aux;
        }
        else
          bipi->setSolFilename(solFile);
      }
      else {
        char *solFile = NULL;
        try {
          if (outDir) {
            solFile = new char[strlen(outDir)+strlen(filename)+6];
            sprintf(solFile, "%s%c%s.sol", outDir, DIRSEP, filename);
          }
          else {
            solFile = new char[strlen(filename)+5];
            sprintf(solFile, "%s.sol", filename);
          }
          bipi->setSolFilename(solFile);
        } catch (exception &e) {
          delete[] solFile;
          throw;
        }
        delete[] solFile;
      }
    } catch (exception &e) {
      delete[] filename;
      throw;
    }
    delete[] filename;
    
    if (threads != -1)
      bipi->setMaxThreads(threads);

    clock_t t, t0;
    t0= clock();
    if (submode == BLOCKIPFORMAT)
      bipi->readBlockIPFormat(BlockIPFormatFile);
    else if (submode == MPS)
      bipi->readMps(mpsFile);
    else if (submode == AMPL)
      bipi->readAmpl(modFile, datFile);
    t= clock();
    double CPUsec= (double) (t-t0)/CLOCKS_PER_SEC;
    cout << "CPU seconds of problem generation: " << CPUsec << endl;
    
    if (solver == BlockIPInterface::BLOCKIP) {
      BlockIP *bip = bipi->getBlockIPInterface();
      double d;
      int i;
      char c;
      for (; argIndex < argc; argIndex += 2) {
        if (!strcmp(argv[argIndex], "-interactive")) {
          interactive = true;
          argIndex -= 1;
        }
        else if (!strcmp(argv[argIndex], "-inf") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_inf(d);
        else if (!strcmp(argv[argIndex], "-m_pw_prec") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i))
          bip->set_m_pw_prec(i);
        else if (!strcmp(argv[argIndex], "-sigma") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_sigma(d);
        else if (!strcmp(argv[argIndex], "-rho") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_rho(d);
        else if (!strcmp(argv[argIndex], "-optim_gap") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_optim_gap(d);
        else if (!strcmp(argv[argIndex], "-optim_pfeas") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_optim_pfeas(d);
        else if (!strcmp(argv[argIndex], "-optim_dfeas") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_optim_dfeas(d);
        else if (!strcmp(argv[argIndex], "-output_freq") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i))
          bip->set_output_freq(i);
        else if (!strcmp(argv[argIndex], "-output") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i))
          bip->set_output((BlockIP::OUTPUT)i);
        else if (!strcmp(argv[argIndex], "-maxiter") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i))
          bip->set_maxiter(i);
        else if (!strcmp(argv[argIndex], "-min_pcgtol") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_min_pcgtol(d);
        else if (!strcmp(argv[argIndex], "-red_pcgtol") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_red_pcgtol(d);
        else if (!strcmp(argv[argIndex], "-init_pcgtol") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_init_pcgtol(d);
        else if (!strcmp(argv[argIndex], "-type_start_point") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i))
          bip->set_type_start_point((BlockIP::TYPE_START_POINT)i);
        else if (!strcmp(argv[argIndex], "-type_direction") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i))
          bip->set_type_direction((BlockIP::TYPE_DIRECTION)i);
        else if (!strcmp(argv[argIndex], "-deactivateLnk") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i))
          bip->set_deactivateLnk(i);
        else if (!strcmp(argv[argIndex], "-iter_pcg") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_maxit_pcg(d);
        else if (!strcmp(argv[argIndex], "-type_reg") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i))
          bip->set_type_reg((BlockIP::TYPE_REG)i);
        else if (!strcmp(argv[argIndex], "-factor_reg") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%lf", &d))
          bip->set_factor_reg(d);
        else if (!strcmp(argv[argIndex], "-show_specrad") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%c", &c))
          bip->set_show_specrad(c=='y');
        else if (!strcmp(argv[argIndex], "-type_comp_dy") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i)) {
          bip->set_type_comp_dy((BlockIP::TYPE_COMP_DY)i);
        }
        else {
          throw 1;
        }
      }
    }
    if (solver == BlockIPInterface::BLOCKIP_MATLAB) {
      double d;
      int i;
      for (; argIndex < argc; argIndex += 2) {
        if (!strcmp(argv[argIndex], "-interactive")) {
          interactive = true;
          argIndex -= 1;
        }
        if (!strcmp(argv[argIndex], "-inf") && ((argIndex+1) < argc)
            && sscanf(argv[argIndex+1], "%lf", &d))
          bipi->infMatlab = d;
        else if (!strcmp(argv[argIndex], "-iter") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i))
          bipi->iterMatlab = i;
        else if (!strcmp(argv[argIndex], "-type_comp_dy") && ((argIndex+1) < argc)
                 && sscanf(argv[argIndex+1], "%i", &i)) {
          if (i == 0) 
            bipi->typeCompMatlab = 1;
          else if (i == 1) 
            bipi->typeCompMatlab = 2;
          else if (i == 2) 
            bipi->typeCompMatlab = 6;
          else if (i == 3) 
            bipi->typeCompMatlab = 7;
        }
      }
    }
    if (argIndex != argc)
    {
      throw 1;
    }

    bool solutionFound = false;
    BlockIPInterface::OPT_STATUS ret = bipi->solve();
    
    switch (ret) {
      case BlockIPInterface::OPTIMAL_SOLUTION:
        cout << "Optimal solution found" << endl;
        solutionFound = true;
        break;
      case BlockIPInterface::TIME_LIMIT_INFEAS:
        cout << "Time limit exhausted with no feasible CTA table" << endl;
        break;
      case BlockIPInterface::TIME_LIMIT_FEAS:
        cout << "Time limit exhausted but a feasible solution has been founded"
        " (likely, not optimal)" << endl;
        solutionFound = true;
        break;
      case BlockIPInterface::FEASIBLE:
        cout << "Feasible solution has been founded (likely, not optimal)" << endl;
        solutionFound = true;
        break;
      case BlockIPInterface::INFEASIBLE:
        cout << "Problem reported as infeasible" << endl;
        break;
      case BlockIPInterface::OTHERWISE:
        cout << "Optimization prematurely stopped without a feasible solution" << endl;
        break;
      case BlockIPInterface::UNBOUNDED:
        cout << "Likely primal or dual unbounded" << endl;
        solutionFound = true;
    }
    if (solutionFound) {
      cout << "Objective function value at actual solution: " << bipi->getObjValue() << endl;
      if (solver == BlockIPInterface::BLOCKIP_MATLAB || solver == BlockIPInterface::BLOCKIP) {
        cout << "Dual objective function: " << bipi->getDualValue() <<
        "\nNumber of IP iterations performed: " << bipi->getIter() <<
        "\nOverall number of PCG iterations performed: " << bipi->getPCGIter() << 
        "\nCPU seconds of optimization: " << bipi->getCPUsec() << endl;
      }
    }
    if (interactive) {
      cout << "Interactive mode, type 'help' for more information\n> " << flush;
      cin.sync();
      do {
        string line;
        getline(cin, line);
        if (line == "exit")
          break;
        stringstream ss;
        ss << line;
        string s1;
        ss >> s1;
        if (s1 == "help") {
          string s2;
          if (ss >> s2) {
            if (s2 == "-v") {
              writeExtHelp();
            }
            else
              cout << "Wrong input" << endl;
          }
          else {
            writeHelp();
          }
        }
        else if (s1 == "pobj")
          cout << bipi->getObjValue() << endl;
        else if (s1 == "dobj")
          cout << bipi->getDualValue() << endl;
        else if (s1 == "it")
          cout << bipi->getIter() << endl;
        else if (s1 == "cgit")
          cout << bipi->getPCGIter() << endl;
        else if (s1 == "x") {
          string s2;
          int i1;
          if (ss >> i1) {
            int i2;
            if (ss >> i2)
              bipi->printX(i1, i2); // x {k} {i}
            else
              bipi->printX(i1); // x {i}
          }
          else {
            ss.clear();
            if (ss >> s2) {
              string s3;
              if (ss >> i1) {
                if (s2 == "slack")
                  bipi->printXSlack(i1); // x slack {i}
                else if (s2 == "block")
                  bipi->printXBlock(i1); // x block {k} 
                else
                  bipi->printX(s2, i1); // x {blockName} {i}
              }
              else {
                ss.clear();
                if (ss >> s3) {
                  if (s2 == "block")
                    bipi->printXBlock(s3); // x block {blockName}
                }
                else {
                  if (s2 == "slack")
                    bipi->printXSlack(); // x slack
                  else
                    bipi->printX(s2); // x {name}
                }
              }
            }
            else
              bipi->printX(); // x
          }
        }
        else if (s1 == "rc") {
          string s2;
          int i1;
          if (ss >> i1) {
            int i2;
            if (ss >> i2)
              bipi->printRC(i1, i2); // rc {k} {i}
            else
              bipi->printRC(i1); // rc {i}
          }
          else {
            ss.clear();
            if (ss >> s2) {
              string s3;
              if (ss >> i1) {
                if (s2 == "slack")
                  bipi->printRCSlack(i1); // rc slack {i}
                else if (s2 == "block")
                  bipi->printRCBlock(i1); // rc block {k} 
                else
                  bipi->printRC(s2, i1); // rc {blockName} {i}
              }
              else {
                ss.clear();
                if (ss >> s3) {
                  if (s2 == "block")
                    bipi->printRCBlock(s3); // rc block {blockName}
                }
                else {
                  if (s2 == "slack")
                    bipi->printRCSlack(); // rc slack
                  else
                    bipi->printRC(s2); // rc {name}
                }
              }
            }
            else
              bipi->printRC(); // x
          }
        }
        else if (s1 == "cons") {
          string s2;
          int i1;
          if (ss >> i1) {
            int i2;
            if (ss >> i2)
              bipi->printCons(i1, i2); // cons {k} {i}
            else
              bipi->printCons(i1); // cons {i}
          }
          else {
            ss.clear();
            if (ss >> s2) {
              string s3;
              if (ss >> i1) {
                if (s2 == "linking")
                  bipi->printConsLinking(i1); // cons linking {i}
                else if (s2 == "block")
                  bipi->printConsBlock(i1); // cons block {k} 
                else
                  bipi->printCons(s2, i1); // cons {blockName} {i}
              }
              else {
                ss.clear();
                if (ss >> s3) {
                  if (s2 == "block")
                    bipi->printConsBlock(s3); // cons block {blockName}
                }
                else {
                  if (s2 == "linking")
                    bipi->printConsLinking(); // cons linking
                  else
                    bipi->printCons(s2); // cons {name}
                }
              }
            }
            else
              bipi->printCons(); // x
          }
        }
        else {
          cout << "Wrong input" << endl;
        }
        cout << "> " << flush;
      } while (true);
    }
  } catch (int &e) {
    delete bipi;
    writeUsageMessage();
    return 1;
  } catch (exception &e) {
    cout << e.what() << endl;
    delete bipi;
    return 1;
  }
  delete bipi;
  if (modFile) {
#ifdef WIN32
    system("rd /S /Q tmp");
#else
    system("/bin/rm -r tmp");
#endif
  }
  return 0;
}

void writeUsageMessage()
{
  cout << "usage: BlockIPPlatform mode [options] [additional options]\n"
  "where mode is:\n"
  "\tIf you want to solve a problem must choose an input format:\n"
  "\t-ampl {model filename} {data filename}\n"
  "\t\tInput problem in sml extended ampl format.\n"
  "\t-mps {mps filename}\n"
  "\t\tInput problem in mps extended to blocks format.\n"
  "\t-BlockIPFormat {BlockIPFormat filename}\n"
  "\t\tInput problem in BlockIP format.\n"
  "\t\t\n"
  "\tIf you want to convert a problem from one input format to other:\n"
  "\t-amplToMps {modelFilename} {dataFilename} {mpsFilename}\n"
  "\t\tConverts a problem in ampl format to mps format\n"
  "\t-amplToBlockIPFormat {modelFilename} {dataFilename} {BlockIPFormatFilename}\n"
  "\t\tConverts a problem in ampl format to BlockIP format\n"
  "\t-mpsToBlockIPFormat {mpsFilename} {BlockIPFormatFilename}\n"
  "\t\tConverts a problem in mps format to BlockIP format\n"
  "\t-BlockIPFormatToMps {BlockIPFormatFilename} {mpsFilename}\n"
  "\t\tConverts a problem in BlockIP format to mps format\n"
  "\n"
  "options for convert mode\n"
  "\t-standarize\n"
  "\t\tStandarize the problem before write the output file converting all\n"
  "\t\trestrictions in equalities and all variables with 0 lower bound.\n"
  "\n"
  "additional options for convert mode\n"
  "\t-outDir {path}\n"
  "\t\tSet the directory where the output files will we written, must exist,\n"
  "\t\tby default it is the actual directory.\n"
  "\t\t\n"
  "options for solve mode\n"
  "\tOptionally you can set some options:\n"
  "\t-solver {solver}\n"
  "\t\tSet the solver to use to solve the problem. Depends of your installation\n"
  "\t\tbut are allowed blockip, blockip_matlab, cplex, xpress, glpk and clp, by default blockip\n"
  "\t\tis used.\n"
  "\t-outDir {path}\n"
  "\t\tSet the directory where the output files will we written, must exist,\n"
  "\t\tby default it is the actual directory.\n"
  "\t-logFile {filename}\n"
  "\t\tSet the name of the log file, by default the same that input file with\n"
  "\t\textension .log.\n"
  "\t-solFile {filename}\n"
  "\t\tSet the name of the solution file, by default the same that input file\n"
  "\t\twith extension .sol.\n"
  "\t-time {number of seconds}\n"
  "\t\tSet the maximum number of seconds that the solver can use to solve the\n"
  "\t\tproblem.\n"
  "\t-threads {number of threads to use}\n"
  "\t\tSet the maximum number of threads that the solver can use to solve the\n"
  "\t\tproblem, by default unlimited.\n"
  "\t-pathMatlab {path}\n"
  "\t\tIf you want to use the Matlab implementation of BlockIP you need prblock_ip.m\n"
  "\t\tfile. The default directory where is expected to find it is \"../prblock_ip\".\n"
  "\t\n"
  "additional for solve mode\n"
  "\tIf the solver to use is BlockIP, you can set addicional options, for the \n"
  "\tdefault values see the constant terms in BlockIP.h. blockip_matlab only \n"
  "\taccepts -in, -iter and -type_comp_dy options \n"
  "\t-interactive\n"
  "\t\tAfter solving the problem starts a interactive mode to retrieve information\n"
  "\t-inf {value}\n"
  "\t\tInfinity value.\n"
  "\t-m_pw_prec {value}\n"
  "\t\tSet number of terms used as preconditioner of the power series expansion of (D-C'*B^{-1}*B)^{-1}\n"
  "\t-sigma {value}\n"
  "\t\tReduction of the centrality parameter at each IP iteration.\n"
  "\t-rho {value}\n"
  "\t\tReduction of the step-length for the primal and dual variables at each \n"
  "\t\t-IP iteration.\n"
  "\t-optim_gap {value}\n"
  "\t\tOptimality gap tolerance.\n"
  "\t-optim_pfeas {value}\n"
  "\t\tPrimal feasibility tolerance.\n"
  "\t-optim_dfeas {value}\n"
  "\t\tDual feasibility tolerance.\n"
  "\t-output_freq {value}\n"
  "\t\tOutput information lines will be printed each selected IP -iterations.\n"
  "\t-output {value}\n"
  "\t\tSet type of output. Can take the values: 0= NONE, 1= SCREEN, 2= FILE, 3=BOTH\n"
  "\t-maxiter {value}\n"
  "\t\tMaximum number of IP iterations.\n"
  "\t-min_pcgtol {value}\n"
  "\t\tMinimum tolerance for the conjugate gradient.\n"
  "\t-red_pcgtol {value}\n"
  "\t\tReduction of pcg_tol at each IP iteration.\n"
  "\t-init_pcgtol {value}\n"
  "\t\tInitial tolerance for the conjugate gradient (by default PCG_TOL_LIN if\n"
  "\t\tlinear problem, PCG_TOL_QUAD if quadratic or -convex- nonlinear).\n"
  "\t-type_start_point {value}\n"
  "\t\tSet how starting point is computed. Can take the values: 0= SIMPLE, 1= QUAD_EQCONS_PROB\n"
  "\t-type_direction {value}\n"
  "\t\tSet which direction is computed. Can take the values: 0= NEWTON, 1= SECOND_ORDER, 2= AUTOMATIC\n"
  "\t-deactivateLnk {value}\n"
  "\t\tSet flag on deactivation of linking\n"
  "\t-iter_pcg {value}\n"
  "\t\tMaximum number of pcg iterations (if 0, then the value will be computed by \n"
  "\t\tthe code).\n"
  "\t-type_reg {value}\n"
  "\t\tSet type of regularization performed. Can take the values: 0=NO_REG, 1= QUAD_REG, 2= PROX_REG\n"
  "\t-factor_reg {value}\n"
  "\t\tSet initial value for regularization term\n"
  "\t-show_specrad {y/n}\n"
  "\t\tWhether estimation of spectral radius by Ritz values is shown at each IP iteration\n"
  "\t-type_comp_dy {value}\n"
  "\t\tScalar that can take values:\n"
  "\t\t0: (default value) the dy direction is computed with Cholesky for blocks,\n"
  "\t\t   and PCG for linking constraints \n"
  "\t\t1: the dy direction is computed with Cholesky for the full A*Theta*A' system.\n"
  "\t\t2: Hybrid Preconditioner: Phase 1) the dy direction is computed with Cholesky\n"
  "\t\t   for blocks, and PCG for linking constraints and Phase 2) dy direction is \n"
  "\t\t   computed with pcg using the split preconditioner for the A*Theta*A' system. \n"
  "\t\t3: Theta0 preconditioner: like type_comp_dy=2, but using Theta0 (Theta of\n"
  "\t\t   slacks of linking constraints) instead of power series preconditioner. \n"
  "\t\t   This may be useful when the space of linking constraints is \"close\" \n"
  "\t\t   (principal angles not large) to the space of block constraints." << endl;
}

void writeHelp()
{
  cout << "help\tShow short help\n"
  "help -v\tShow extended help\n"
  "pobj\tShow primal objective function\n"
  "dobj\tShow dual objective function\n"
  "it\tShow the number of IP iterations performed\n"
  "cgit\tShow the overall number of PCG iterations performed\n"
  "exit\tEnds the interactive mode\n"
  "\n"
  "x\tDisplay the primal variables value of blocks and slacks\n"
  "x {i}\tExample: x 5\n"
  "x {name}\tExample: x Block1:Var5\n"
  "x {k} {i}\tExample: x 1 5\n"
  "x {blockName} {i}\tExample: x Block1 5\n"
  "x block {k}\tExample: x block 1\n"
  "x block {blockName}\tExample: x block Block1\n"
  "x slack\tExample: x slack\n"
  "x slack {i}\tExample: x slack 3\n"
  "\n"
  "rc\tDisplay the reduced cost of primal variables of blocks and slacks\n"
  "rc {i}\tExample: rc 5\n"
  "rc {name}\tExample: rc Block1:Var5\n"
  "rc {k} {i}\tExample: rc 1 5\n"
  "rc {blockName} {i}\tExample: rc Block1 5\n"
  "rc block {k}\tExample: rc block 1\n"
  "rc block {blockName}\tExample: rc block Block1\n"
  "rc slack\tExample: rc slack\n"
  "rc slack {i}\tExample: rc slack 3\n"
  "\n"
  "cons\tDisplay the dual variables value of blocks and linking constraints\n"
  "cons {i}\tExample: cons 5\n"
  "cons {name}\tExample: cons Block1:Cons5\n"
  "cons {k} {i}\tExample: cons 1 5\n"
  "cons {blockName} {i}\tExample: cons Block1 5\n"
  "cons block {k}\tExample: cons block 1\n"
  "cons block {blockName}\tExample: cons block Block1\n"
  "cons linking\tExample: cons linking\n"
  "cons linking {i}\tExample: cons linking 3"  << endl;
}

void writeExtHelp()
{
  cout << "help\tShow short help\n"
  "help -v\tShow extended help\n"
  "pobj\tShow primal objective function\n"
  "dobj\tShow dual objective function\n"
  "it\tShow the number of IP iterations performed\n"
  "cgit\tShow the overall number of PCG iterations performed\n"
  "exit\tEnds the interactive mode\n"
  "\n"
  "Indices start in 0.\n"
  "Display the primal variables value of blocks and slacks\n"
  "x\n"
  "\tShow all the primal variables of blocks and slacks\n"
  "x {i}\tExample: x 5\n"
  "\tShow the value of the i-th primal variable\n"
  "x {name}\tExample: x Block1:Var5\n"
  "\tShow the value of the named primal variable\n"
  "x {k} {i}\tExample: x 1 5\n"
  "\tShow the value of the i-th primal variable inside the block k\n"
  "x {blockName} {i}\tExample: x Block1 5\n"
  "\tShow the value of the i-th primal variable inside the named block\n"
  "x block {k}\tExample: x block 1\n"
  "\tShow the value of all primal variables inside the block k\n"
  "x block {blockName}\tExample: x block Block1\n"
  "\tShow the value of all primal variables inside the named block\n"
  "x slack\tExample: x slack\n"
  "\tShow the value of all primal variables inside the slacks\n"
  "x slack {i}\tExample: x slack 3\n"
  "\tShow the value of the i-th primal variable inside the slacks\n"
  "\n"
  "Display the reduced cost of primal variables of blocks and slacks\n"
  "rc\n"
  "\tShow all the reduced cost of primal variables of blocks and slacks\n"
  "rc {i}\tExample: rc 5\n"
  "\tShow the reduced cost of the i-th primal variable\n"
  "rc {name}\tExample: rc Block1:Var5\n"
  "\tShow the reduced cost of the named primal variable\n"
  "rc {k} {i}\tExample: rc 1 5\n"
  "\tShow the reduced cost of the i-th primal variable inside the block k\n"
  "rc {blockName} {i}\tExample: rc Block1 5\n"
  "\tShow the reduced cost of the i-th primal variable inside the named block\n"
  "rc block {k}\tExample: rc block 1\n"
  "\tShow the reduced cost of all primal variables inside the block k\n"
  "rc block {blockName}\tExample: rc block Block1\n"
  "\tShow the reduced cost of all primal variables inside the named block\n"
  "rc slack\tExample: rc slack\n"
  "\tShow the reduced cost of all primal variables inside the slacks\n"
  "rc slack {i}\tExample: rc slack 3\n"
  "\tShow the reduced cost of the i-th primal variable inside the slacks\n"
  "\n"
  "Display the dual variables value of blocks and linking constraints\n"
  "cons\n"
  "\tShow all the dual variables of blocks and linking constraints\n"
  "cons {i}\tExample: cons 5\n"
  "\tShow the dual variable of the i-th constraint\n"
  "cons {name}\tExample: cons Block1:Cons5\n"
  "\tShow the dual variable of the named constraint\n"
  "cons {k} {i}\tExample: cons 1 5\n"
  "\tShow the dual variable of the i-th constraint inside the block k\n"
  "cons {blockName} {i}\tExample: cons Block1 5\n"
  "\tShow the dual variable of the i-th constraint inside the named block\n"
  "cons block {k}\tExample: cons block 1\n"
  "\tShow all the dual variables inside the block k\n"
  "cons block {blockName}\tExample: cons block Block1\n"
  "\tShow all the dual varialbes inside the named block\n"
  "cons linking\tExample: cons linking\n"
  "\tShow all the dual varialbes inside the linking constraints\n"
  "cons linking {i}\tExample: cons linking 3\n"
  "\tShow the dual variable of the i-th constraint inside the linking constraints" << endl;
}
